from mnspy.ecuaciones_diferenciales_parciales.mef import Nodo, Elemento, Rigidez
from mnspy.utilidades import es_notebook, _generar_matrix
from IPython.display import display, Math
from tabulate import tabulate
import numpy as np
import matplotlib.pyplot as plt

from matplotlib.path import Path
import matplotlib.patches as patches

plt.rcParams.update(plt.rcParamsDefault)


class Armadura(Elemento):
    """Representa un elemento de armadura (truss) en 2D.

    Este elemento solo soporta cargas axiales (tensión y compresión) y se utiliza
    para modelar estructuras de armaduras planas. Cada nodo tiene dos grados de
    libertad (desplazamiento en 'x' y 'y').

    Attributes
    ----------
    A : float
        Área de la sección transversal del elemento.
    E : float
        Módulo de Young del material.
    _c : float
        Coseno del ángulo del elemento con respecto al eje x global.
    _s : float
        Seno del ángulo del elemento con respecto al eje x global.
    _kp : np.ndarray
        Matriz de rigidez local del elemento.
    _T : np.ndarray
        Matriz de transformación de coordenadas locales a globales.
    """
    def __init__(self, nombre: str, nodo_i: Nodo, nodo_j: Nodo, A: float, E: float):
        """Constructor para el elemento de armadura."""
        super().__init__(nombre, nodo_i, nodo_j)
        c = (self._nodo_j.punto[0] - self._nodo_i.punto[0]) / self._L
        s = (self._nodo_j.punto[1] - self._nodo_i.punto[1]) / self._L
        self._c = c
        self._s = s
        self._A = A
        self._E = E
        self._kp = np.array([[1, 0, -1, 0],
                             [0, 0, 0, 0],
                             [-1, 0, 1, 0],
                             [0, 0, 0, 0]],
                            dtype=np.double) * self._A * self._E / self._L
        self._T = np.array([[c, s, 0, 0],
                            [-s, c, 0, 0],
                            [0, 0, c, s],
                            [0, 0, -s, c]], dtype=np.double)
        self._k = Rigidez(np.matmul(np.matmul(self._T.transpose(), self._kp), self._T), [self._nodo_i, self._nodo_j],
                          ['x', 'y'])
        self._fuerzas_i = np.zeros((len(self._k.grados), 1))
        self._fuerzas_j = np.zeros((len(self._k.grados), 1))
        self._nodo_i.grados_libertad['x'].label_reaccion = 'F'
        self._nodo_i.grados_libertad['x'].label_fuerza = 'f'
        self._nodo_i.grados_libertad['y'].label_reaccion = 'F'
        self._nodo_i.grados_libertad['y'].label_fuerza = 'f'
        self._nodo_j.grados_libertad['x'].label_reaccion = 'F'
        self._nodo_j.grados_libertad['x'].label_fuerza = 'f'
        self._nodo_j.grados_libertad['y'].label_reaccion = 'F'
        self._nodo_j.grados_libertad['y'].label_fuerza = 'f'

    def _repr_latex_(self):
        vec_d = np.array(self._k.obtener_etiquetas_desplazamientos(), dtype=object).reshape(-1,1)
        vec_f = np.array(self._obtener_etiquetas_fuerzas(), dtype=object).reshape(-1,1)
        texto_latex = r'\left\{\begin{array}{c}'
        texto_latex += _generar_matrix(vec_f,
                                       '{:}') + r'\end{array}\right\}_{\{f\}}=\left[\begin{array}{' + 'c' * \
                       self._k.obtener_matriz().shape[1] + '}'
        texto_latex += _generar_matrix(self._k.obtener_matriz(),
                                       '{:G}') + r'\end{array}\right]_{[k]}\cdot\left\{\begin{array}{c}'
        texto_latex += _generar_matrix(vec_d, '{:}') + r'\end{array}\right\}_{\{d\}}'
        return '$' + texto_latex + '$'

    def __repr__(self):
        """Overload del método __repr__

        Returns
        -------
        Muestra el nombre del elemento
        """
        # self.mostrar_sistema()
        return 'Armadura: ' + self.nombre

    def __str__(self):
        """Overload del método __str__
        Returns
        -------
        Información del elemento
        """
        return 'Armadura: ' + self.nombre

    def _obtener_valor_fuerza(self):
        return np.matmul(np.array([[-self._c, -self._s, self._c, self._s]]), self._k.obtener_desplazamientos())[
            0, 0] * self._E * self._A / self._L

    def get_matriz_rigidez_local(self):
        return self._kp

    def get_matriz_T(self):
        return self._T

    def fuerza(self):
        if es_notebook():
            indice = [r'$f^{(' + self.nombre + r')}$']
            return tabulate({'Fuerza': [self._obtener_valor_fuerza()]}, headers='keys', showindex=indice,
                            tablefmt='html')
        else:
            return self._obtener_valor_fuerza()

    def esfuerzo(self):
        if es_notebook():
            indice = [r'$\sigma^{(' + self.nombre + r')}$']
            return tabulate({'Esfuerzo': [self._obtener_valor_fuerza() / self._A]}, headers='keys', showindex=indice,
                            tablefmt='html')
        else:
            return self._obtener_valor_fuerza() / self._A

    def mostrar_sistema(self, reducida: bool = False):
        vec_d = np.array(self._k.obtener_etiquetas_desplazamientos(reducida), dtype=object).reshape(-1,1)
        vec_f = np.array(self._obtener_etiquetas_fuerzas(reducida), dtype=object).reshape(-1,1)
        if es_notebook():
            texto_latex = r'\left\{\begin{array}{c}'
            texto_latex += _generar_matrix(vec_f,
                                           '{:}') + r'\end{array}\right\}_{\{f\}}=\left[\begin{array}{' + 'c' * \
                           self._k.obtener_matriz(reducida).shape[1] + '}'
            texto_latex += _generar_matrix(self._k.obtener_matriz(reducida),
                                           '{:G}') + r'\end{array}\right]_{[k]}\cdot\left\{\begin{array}{c}'
            texto_latex += _generar_matrix(vec_d, '{:}') + r'\end{array}\right\}_{\{d\}}'
            display(Math(texto_latex))
        else:
            return np.array2string(self._k.obtener_matriz(reducida),
                                   formatter={'float_kind': lambda x: '{:}'.format(x)})

    def fuerzas_internas(self):
        fuerzas_internas = np.matmul(self._k.k, self._obtener_desplazamientos())
        if es_notebook():
            indice = ['$' + label + '$' for label in self._obtener_etiquetas_fuerzas()]
            return tabulate({'Fuerzas internas': fuerzas_internas}, headers='keys', showindex=indice, tablefmt='html')
        else:
            print(fuerzas_internas)

    def fuerzas_internas_local(self):
        fuerzas_internas = np.matmul(self._T[[0, 2], :], np.matmul(self._k.k, self._obtener_desplazamientos()))
        if es_notebook():
            indice = ['$' + label + '$' for label in self._obtener_etiquetas_fuerzas_local()]
            return tabulate({'Fuerzas internas local': fuerzas_internas}, headers='keys', showindex=indice,
                            tablefmt='html')
        else:
            print(fuerzas_internas)

    def _obtener_etiquetas_fuerzas(self, reducida: bool = False):
        etq = [nodo.grados_libertad[gl].label_fuerza + '^{(' + self.nombre + ')}_{' + nodo.nombre + gl + '}' for nodo
               in [self._nodo_i, self._nodo_j] for gl in self._k.grados if
               not reducida or nodo.grados_libertad[gl].valor]
        return etq

    def _obtener_etiquetas_fuerzas_local(self, reducida: bool = False):
        etq = [nodo.grados_libertad['x'].label_fuerza + r'^{\prime(' + self.nombre + ')}_{' + nodo.nombre + 'x}' for
               nodo in [self._nodo_i, self._nodo_j] if not reducida or nodo.grados_libertad['x'].valor]
        return etq

    def _obtener_desplazamientos(self) -> np.ndarray:
        desplazamiento = None
        for item in self._k.lista_nodos:
            d_i = [d.desplazamiento for d in item.grados_libertad.values() if d.gl in self._k.grados]
            if desplazamiento is None:
                desplazamiento = d_i
            else:
                desplazamiento = np.hstack((desplazamiento, d_i))
        return np.array(desplazamiento).reshape(-1,1)

    def __obtener_path_elemento(self, espesor: np.double) -> Path:
        """Genera la ruta SVG para dibujar el elemento de armadura."""
        x_1, y_1, Z_1 = self._nodo_i.punto
        x_2, y_2, Z_2 = self._nodo_j.punto
        c = self._c
        s = self._s
        t = espesor

        vertices = np.array([(x_1 + 0.5 * t * s, y_1 - 0.5 * t * c), (x_2 + 0.5 * t * s, y_2 - 0.5 * t * c),
                             (x_2 + 0.5 * t * s + 0.7 * t * c, y_2 - 0.5 * t * c + 0.7 * t * s),
                             (x_2 - 0.5 * t * s + 0.7 * t * c, y_2 + 0.5 * t * c + 0.7 * t * s),
                             (x_2 - 0.5 * t * s, y_2 + 0.5 * t * c), (x_1 - 0.5 * t * s, y_1 + 0.5 * t * c),
                             (x_1 - 0.5 * t * s - 0.7 * t * c, y_1 + 0.5 * t * c - 0.7 * t * s),
                             (x_1 + 0.5 * t * s - 0.7 * t * c, y_1 - 0.5 * t * c - 0.7 * t * s),
                             (x_1 + 0.5 * t * s, y_1 - 0.5 * t * c)])
        codes = [Path.MOVETO, Path.LINETO, Path.CURVE4, Path.CURVE4, Path.CURVE4, Path.LINETO, Path.CURVE4, Path.CURVE4,
                 Path.CURVE4]
        return Path(vertices, codes)

    def diagrama_fuerzas_internas(self):
        """Dibuja un diagrama de cuerpo libre del elemento con sus fuerzas internas globales."""
        fuerzas_internas = np.matmul(self._k.k, self._obtener_desplazamientos())
        etq = self._obtener_etiquetas_fuerzas()
        fig, ax = plt.subplots()
        ax.add_patch(patches.PathPatch(self.__obtener_path_elemento(self._L / 20), edgecolor='royalblue',
                                       facecolor='lightsteelblue', lw=0.2))
        fuerza = fuerzas_internas[0, 0]
        if fuerza <= 0.0:
            pos = -50
            ali = 'right'
            sentido = '<-'
        else:
            pos = -50
            ali = 'right'
            sentido = '->'
        ax.annotate('$' + etq[0] + '=' + '{:G}$'.format(abs(fuerza)),
                    xy=(self._nodo_i.punto[0:2]), xycoords='data',
                    xytext=(pos, 0), textcoords='offset points', va='center', ha=ali, size=12,
                    arrowprops=dict(arrowstyle=sentido, connectionstyle="arc3"))
        fuerza = fuerzas_internas[1, 0]
        if fuerza <= 0.0:
            pos = -50
            ali = 'top'
            sentido = '<-'
        else:
            pos = -50
            ali = 'top'
            sentido = '->'
        ax.annotate('$' + etq[1] + '=' + '{:G}$'.format(abs(fuerza)),
                    xy=(self._nodo_i.punto[0:2]), xycoords='data',
                    xytext=(0, pos), textcoords='offset points', va=ali, ha='center', size=12,
                    arrowprops=dict(arrowstyle=sentido, connectionstyle="arc3"))
        fuerza = fuerzas_internas[2, 0]
        if fuerza <= 0.0:
            pos = 50
            ali = 'left'
            sentido = '->'
        else:
            pos = 50
            ali = 'left'
            sentido = '<-'
        ax.annotate('$' + etq[2] + '=' + '{:G}$'.format(abs(fuerza)),
                    xy=(self._nodo_j.punto[0:2]), xycoords='data',
                    xytext=(pos, 0), textcoords='offset points', va='center', ha=ali, size=12,
                    arrowprops=dict(arrowstyle=sentido, connectionstyle="arc3"))
        fuerza = fuerzas_internas[3, 0]
        if fuerza <= 0.0:
            pos = 50
            ali = 'bottom'
            sentido = '->'
        else:
            pos = 50
            ali = 'bottom'
            sentido = '<-'
        ax.annotate('$' + etq[3] + '=' + '{:G}$'.format(abs(fuerza)),
                    xy=(self._nodo_j.punto[0:2]), xycoords='data',
                    xytext=(0, pos), textcoords='offset points', va=ali, ha='center', size=12,
                    arrowprops=dict(arrowstyle=sentido, connectionstyle="arc3"))
        plt.scatter([self._nodo_i.punto[0], self._nodo_j.punto[0]], [self._nodo_i.punto[1], self._nodo_j.punto[1]],
                    c='navy',
                    marker='o')

        ax.axis('equal')
        ax.axis('off')
        plt.show()

    def diagrama_fuerzas_internas_local(self):
        """Dibuja un diagrama de cuerpo libre con las fuerzas internas locales (axiales)."""
        fuerzas_internas = np.matmul(self._T[[0, 2], :], np.matmul(self._k.k, self._obtener_desplazamientos()))
        etq = self._obtener_etiquetas_fuerzas_local()
        fig, ax = plt.subplots()
        ax.add_patch(patches.PathPatch(self.__obtener_path_elemento(self._L / 20), edgecolor='royalblue',
                                       facecolor='lightsteelblue', lw=0.2))
        ang = np.degrees(np.arccos(self.get_coseno())) * np.sign(self.get_seno())
        fuerza = fuerzas_internas[0, 0]
        if fuerza <= 0.0:
            pos_x = -50 * self._c
            pos_y = -50 * self._s
            ali_v = 'center'
            ali_h = 'right'
            sentido = '<-'
        else:
            pos_x = -50 * self._c
            pos_y = -50 * self._s
            ali_v = 'center'
            ali_h = 'right'
            sentido = '->'
        ax.annotate('',
                    xy=(self._nodo_i.punto[0:2]), xycoords='data',
                    xytext=(pos_x, pos_y), textcoords='offset points', va=ali_v, ha=ali_h, size=12,
                    arrowprops=dict(arrowstyle=sentido, connectionstyle="arc3"))
        ax.annotate('$' + etq[0] + '=' + '{:G}$'.format(abs(fuerza)),
                    xy=(self._nodo_i.punto[0:2]), xycoords='data',
                    xytext=(pos_x, pos_y), textcoords='offset points', va=ali_v, ha=ali_h, size=12, rotation=ang,
                    rotation_mode='anchor')
        fuerza = fuerzas_internas[1, 0]
        if fuerza <= 0.0:
            pos_x = 50 * self._c
            pos_y = 50 * self._s
            ali_v = 'center'
            ali_h = 'left'
            sentido = '->'
        else:
            pos_x = 50 * self._c
            pos_y = 50 * self._s
            ali_v = 'center'
            ali_h = 'left'
            sentido = '<-'
        ax.annotate('',
                    xy=(self._nodo_j.punto[0:2]), xycoords='data',
                    xytext=(pos_x, pos_y), textcoords='offset points', va=ali_v, ha=ali_h, size=12,
                    arrowprops=dict(arrowstyle=sentido, connectionstyle="arc3"))
        ax.annotate('$' + etq[1] + '=' + '{:G}$'.format(abs(fuerza)),
                    xy=(self._nodo_j.punto[0:2]), xycoords='data',
                    xytext=(pos_x, pos_y), textcoords='offset points', va=ali_v, ha=ali_h, size=12, rotation=ang,
                    rotation_mode='anchor')
        plt.scatter([self._nodo_i.punto[0], self._nodo_j.punto[0]], [self._nodo_i.punto[1], self._nodo_j.punto[1]],
                    c='navy',
                    marker='o')

        ax.axis('equal')
        ax.axis('off')
        plt.show()



from mnspy.ecuaciones_diferenciales_parciales.mef import Nodo, Elemento, Rigidez
from mnspy.utilidades import es_notebook, _generar_matrix
from mnspy.ecuaciones_algebraicas_lineales import Gauss
from IPython.display import display, Math
from tabulate import tabulate
import numpy as np
import matplotlib.pyplot as plt

from matplotlib.path import Path
import matplotlib.patches as patches
import matplotlib.transforms as transforms
import sympy as sp

TOL_CERO = 1E-10
sp.init_printing(use_latex=True)

# plt.rcParams.update({
#     "text.usetex": True,
#     "font.family": "sans-serif",
#     "font.sans-serif": "Helvetica",
# })
plt.rcParams.update(plt.rcParamsDefault)


class Marco(Elemento):
    """Representa un elemento de marco plano (frame) en 2D.

    Este elemento combina el comportamiento de una viga y una armadura, ya que
    puede soportar cargas axiales, cortantes y momentos flectores. Se utiliza
    para modelar estructuras como pórticos y marcos rígidos.

    Cada nodo del elemento tiene tres grados de libertad:
    1. Desplazamiento en la dirección 'x' global.
    2. Desplazamiento en la dirección 'y' global.
    3. Rotación alrededor del eje 'z' global.

    El elemento puede tener una rótula (articulación) en el nodo final (`nodo_j`),
    lo que significa que no transmite momento en ese punto.

    Attributes
    ----------
    A : float
        Área de la sección transversal del elemento.
    E : float
        Módulo de Young del material.
    I : float
        Momento de inercia de la sección transversal.
    _c : float
        Coseno del ángulo del elemento con respecto al eje x global.
    _s : float
        Seno del ángulo del elemento con respecto al eje x global.
    _kp : np.ndarray
        Matriz de rigidez local del elemento (6x6).
    _T : np.ndarray
        Matriz de transformación de coordenadas (6x6).
    _k : Rigidez
        Objeto `Rigidez` que contiene la matriz de rigidez global del elemento.
    _cargas : dict
        Diccionario que almacena las cargas aplicadas sobre el elemento.
    """

    def __init__(self, nombre: str, nodo_i: Nodo, nodo_j: Nodo, A: float, E: float, I: float):
        """Constructor para el elemento de marco plano.

        Parameters
        ----------
        nombre : str
            Nombre o identificador del elemento.
        nodo_i : Nodo
            Nodo inicial del elemento.
        nodo_j : Nodo
            Nodo final del elemento. Puede ser una rótula si `nodo_j.es_rotula` es True.
        A : float
            Área de la sección transversal.
        E : float
            Módulo de Young del material.
        I : float
            Momento de inercia de la sección transversal.
        """
        super().__init__(nombre, nodo_i, nodo_j)
        c = (self._nodo_j.punto[0] - self._nodo_i.punto[0]) / self._L
        s = (self._nodo_j.punto[1] - self._nodo_i.punto[1]) / self._L
        self._c = c
        self._s = s
        self._A = A
        self._E = E

        # --- Cálculo de la matriz de rigidez local [kp] ---
        # Constantes para las fórmulas de rigidez de viga y barra
        c_1 = A * E / self._L  # Rigidez axial (AE/L)
        c_2 = E * I / self._L ** 3  # Constante de rigidez a flexión (EI/L^3)

        # Se selecciona la matriz de rigidez local apropiada
        # si el nodo final es una rótula (articulación).
        if nodo_j.es_rotula:
            # Matriz de rigidez local para un marco con rótula en el extremo j (articulado-empotrado)
            self._kp = np.array([[c_1, 0, 0, -c_1, 0, 0],
                                 [0, 3 * c_2, 3 * c_2 * self._L, 0, -3 * c_2, 0],
                                 [0, 3 * c_2 * self._L, 3 * c_2 * self._L ** 2, 0, -3 * c_2 * self._L, 0],
                                 [-c_1, 0, 0, c_1, 0, 0],
                                 [0, -3 * c_2, -3 * c_2 * self._L, 0, 3 * c_2, 0],
                                 [0, 0, 0, 0, 0, 0]],
                                dtype=np.double)
        else:
            # Matriz de rigidez local para un marco estándar (empotrado-empotrado)
            self._kp = np.array([[c_1, 0, 0, -c_1, 0, 0],
                                 [0, 12 * c_2, 6 * c_2 * self._L, 0, -12 * c_2, 6 * c_2 * self._L],
                                 [0, 6 * c_2 * self._L, 4 * c_2 * self._L ** 2, 0, -6 * c_2 * self._L,
                                  2 * c_2 * self._L ** 2],
                                 [-c_1, 0, 0, c_1, 0, 0],
                                 [0, -12 * c_2, -6 * c_2 * self._L, 0, 12 * c_2, -6 * c_2 * self._L],
                                 [0, 6 * c_2 * self._L, 2 * c_2 * self._L ** 2, 0, -6 * c_2 * self._L,
                                  4 * c_2 * self._L ** 2]],
                                dtype=np.double)

        # --- Matriz de transformación de coordenadas [T] ---
        self._T = np.array([[c, s, 0, 0, 0, 0],
                            [-s, c, 0, 0, 0, 0],
                            [0, 0, 1, 0, 0, 0],
                            [0, 0, 0, c, s, 0],
                            [0, 0, 0, -s, c, 0],
                            [0, 0, 0, 0, 0, 1]], dtype=np.double)

        # --- Matriz de rigidez en coordenadas globales [k] = [T]^T * [kp] * [T] ---
        self._k = Rigidez(np.matmul(np.matmul(self._T.transpose(), self._kp), self._T), [self._nodo_i, self._nodo_j],
                          ['x', 'y', 'eje_z'])

        # Inicialización de vectores de fuerzas nodales equivalentes
        self._fuerzas_i = np.zeros((len(self._k.grados), 1))
        self._fuerzas_j = np.zeros((len(self._k.grados), 1))
        # Fuerzas de corrección por rótula
        self._fuerzas_i_rot = np.zeros((len(self._k.grados), 1))
        self._fuerzas_j_rot = np.zeros((len(self._k.grados), 1))

        # Asignación de etiquetas para fuerzas y reacciones
        self._nodo_i.grados_libertad['x'].label_reaccion = 'F'
        self._nodo_i.grados_libertad['x'].label_fuerza = 'f'
        self._nodo_i.grados_libertad['y'].label_reaccion = 'F'
        self._nodo_i.grados_libertad['y'].label_fuerza = 'f'
        self._nodo_i.grados_libertad['eje_z'].label_reaccion = 'M'
        self._nodo_i.grados_libertad['eje_z'].label_fuerza = 'm'
        self._nodo_j.grados_libertad['x'].label_reaccion = 'F'
        self._nodo_j.grados_libertad['x'].label_fuerza = 'f'
        self._nodo_j.grados_libertad['y'].label_reaccion = 'F'
        self._nodo_j.grados_libertad['y'].label_fuerza = 'f'
        self._nodo_j.grados_libertad['eje_z'].label_reaccion = 'M'
        self._nodo_j.grados_libertad['eje_z'].label_fuerza = 'm'

        # Diccionario para almacenar cargas en el elemento
        self._cargas = {'distribuida': [], 'puntual': [], 'axial': [],
                        'momento': []}
        # Diccionario para coeficientes de ecuaciones de diagramas
        self._coef = {'A': 0.0, 'B': 0.0, 'c_1': 0.0, 'c_2': 0.0, 'c_3': 0.0, 'c_4': 0.0}
        self._A = A
        self._E = E
        self._I = I

    def _repr_latex_(self):
        """Representación en LaTeX del sistema local del elemento para notebooks."""
        vec_d = np.array(self._k.obtener_etiquetas_desplazamientos(), dtype=object).reshape(-1, 1)
        vec_f = np.array(self._obtener_etiquetas_fuerzas(), dtype=object).reshape(-1, 1)
        texto_latex = r'\left\{\begin{array}{c}'
        texto_latex += _generar_matrix(vec_f,
                                       '{:}') + r'\end{array}\right\}_{\{f\}}=\left[\begin{array}{' + 'c' * \
                       self._k.obtener_matriz().shape[1] + '}'
        texto_latex += _generar_matrix(self._k.obtener_matriz(),
                                       '{:G}') + r'\end{array}\right]_{[k]}\cdot\left\{\begin{array}{c}'
        texto_latex += _generar_matrix(vec_d, '{:}') + r'\end{array}\right\}_{\{d\}}-\left\{\begin{array}{c}'
        texto_latex += _generar_matrix(self._obtener_fuerzas(),
                                       '{:}') + r'\end{array}\right\}_{\{f_{o}\}}'
        if self._nodo_j.es_rotula:
            texto_latex += r'-\left\{\begin{array}{c}' + _generar_matrix(self._obtener_fuerzas_por_rotula(),
                                                                         '{:}') + r'\end{array}\right\}_{\{f_{rot}\}}'
        return '$' + texto_latex + '$'

    def __repr__(self):
        """Representación del objeto como string."""
        return 'Marco: ' + self.nombre

    def __str__(self):
        """Representación del objeto como string."""
        return 'Marco: ' + self.nombre

    def agregar_carga_distribuida(self, q: float):
        """Agrega una carga uniformemente distribuida perpendicular al elemento.

        Calcula las fuerzas y momentos de empotramiento perfecto (fixed-end actions)
        correspondientes y los suma a los vectores de fuerza nodal del elemento.
        Si el nodo final es una rótula, se aplica una corrección para liberar
        el momento.

        Parameters
        ----------
        q : float
            Magnitud de la carga distribuida. Un valor positivo indica que la carga
            actúa en la dirección positiva del eje 'y' local del elemento.
        """
        # Fuerzas y momentos de empotramiento perfecto para carga distribuida
        self._nodo_i.grados_libertad['x'].fuerza += -q * self._L * self._s / 2
        self._fuerzas_i[0, 0] += -q * self._L * self._s / 2
        self._nodo_i.grados_libertad['y'].fuerza += q * self._L * self._c / 2
        self._fuerzas_i[1, 0] += q * self._L * self._c / 2
        self._nodo_i.grados_libertad['eje_z'].fuerza += q * self._L ** 2 / 12
        self._fuerzas_i[2, 0] += q * self._L ** 2 / 12
        self._nodo_j.grados_libertad['x'].fuerza += -q * self._L * self._s / 2
        self._fuerzas_j[0, 0] += -q * self._L * self._s / 2
        self._nodo_j.grados_libertad['y'].fuerza += q * self._L * self._c / 2
        self._fuerzas_j[1, 0] += q * self._L * self._c / 2
        self._nodo_j.grados_libertad['eje_z'].fuerza += -q * self._L ** 2 / 12
        self._fuerzas_j[2, 0] += -q * self._L ** 2 / 12

        self._cargas['distribuida'].append([q, q])

        # Si el nodo j es una rótula, se debe liberar el momento en ese extremo.
        # Esto se hace aplicando un momento corrector en el nodo i y las fuerzas
        # cortantes correspondientes en ambos nodos para mantener el equilibrio.
        if self._nodo_j.es_rotula:
            self._nodo_i.grados_libertad['x'].fuerza -= -self._s * (-q * self._L ** 2 / 12) * 3.0 / (2.0 * self._L)
            self._fuerzas_i_rot[0, 0] -= -self._s * (-q * self._L ** 2 / 12) * 3.0 / (2.0 * self._L)
            self._nodo_i.grados_libertad['y'].fuerza -= self._c * (-q * self._L ** 2 / 12) * 3.0 / (2.0 * self._L)
            self._fuerzas_i_rot[1, 0] -= self._c * (-q * self._L ** 2 / 12) * 3.0 / (2.0 * self._L)
            self._nodo_i.grados_libertad['eje_z'].fuerza -= 0.5 * (-q * self._L ** 2 / 12)
            self._fuerzas_i_rot[2, 0] -= 0.5 * (-q * self._L ** 2 / 12)
            self._nodo_j.grados_libertad['x'].fuerza -= -self._s * -(-q * self._L ** 2 / 12) * 3.0 / (2.0 * self._L)
            self._fuerzas_j_rot[0, 0] -= -self._s * -(-q * self._L ** 2 / 12) * 3.0 / (2.0 * self._L)
            self._nodo_j.grados_libertad['y'].fuerza -= self._c * -(-q * self._L ** 2 / 12) * 3.0 / (2.0 * self._L)
            self._fuerzas_j_rot[1, 0] -= self._c * -(-q * self._L ** 2 / 12) * 3.0 / (2.0 * self._L)
            self._nodo_j.grados_libertad['eje_z'].fuerza -= -q * self._L ** 2 / 12
            self._fuerzas_j_rot[2, 0] -= -q * self._L ** 2 / 12

    def agregar_carga_trapezoidal(self, q_1: float, q_2: float):
        """Agrega una carga trapezoidalmente distribuida perpendicular al elemento.

        Calcula las fuerzas y momentos de empotramiento perfecto (fixed-end actions)
        correspondientes y los suma a los vectores de fuerza nodal del elemento.
        Si el nodo final es una rótula, se aplica una corrección.

        Parameters
        ----------
        q_1 : float
            Magnitud de la carga en el nodo inicial (i).
        q_2 : float
            Magnitud de la carga en el nodo final (j).
        """
        self._nodo_i.grados_libertad['x'].fuerza += -q_1 * self._L * self._s / 2 - 3 * (
                q_2 - q_1) * self._L * self._s / 20
        self._fuerzas_i[0, 0] += -q_1 * self._L * self._s / 2 - 3 * (q_2 - q_1) * self._L * self._s / 20
        self._nodo_i.grados_libertad['y'].fuerza += q_1 * self._L * self._c / 2 + 3 * (
                q_2 - q_1) * self._L * self._c / 20
        self._fuerzas_i[1, 0] += q_1 * self._L * self._c / 2 + 3 * (q_2 - q_1) * self._L * self._c / 20
        self._nodo_i.grados_libertad['eje_z'].fuerza += q_1 * self._L ** 2 / 12 + (q_2 - q_1) * self._L ** 2 / 30
        self._fuerzas_i[2, 0] += q_1 * self._L ** 2 / 12 + (q_2 - q_1) * self._L ** 2 / 30
        self._nodo_j.grados_libertad['x'].fuerza += -q_1 * self._L * self._s / 2 - 7 * (
                q_2 - q_1) * self._L * self._s / 20
        self._fuerzas_j[0, 0] += -q_1 * self._L * self._s / 2 - 7 * (q_2 - q_1) * self._L * self._s / 20
        self._nodo_j.grados_libertad['y'].fuerza += q_1 * self._L * self._c / 2 + 7 * (
                q_2 - q_1) * self._L * self._c / 20
        self._fuerzas_j[1, 0] += q_1 * self._L * self._c / 2 + 7 * (q_2 - q_1) * self._L * self._c / 20
        self._nodo_j.grados_libertad['eje_z'].fuerza += -q_1 * self._L ** 2 / 12 - (q_2 - q_1) * self._L ** 2 / 20
        self._fuerzas_j[2, 0] += -q_1 * self._L ** 2 / 12 - (q_2 - q_1) * self._L ** 2 / 20

        self._cargas['distribuida'].append([q_1, q_2])

        # Corrección por rótula en el nodo j
        if self._nodo_j.es_rotula:
            self._nodo_i.grados_libertad['x'].fuerza -= -self._s * (-q_1 * self._L ** 2 / 12 - (
                    q_2 - q_1) * self._L ** 2 / 20) * 3.0 / (2.0 * self._L)
            self._fuerzas_i_rot[0, 0] -= -self._s * (
                    -q_1 * self._L ** 2 / 12 - (q_2 - q_1) * self._L ** 2 / 20) * 3.0 / (
                                                 2.0 * self._L)
            self._nodo_i.grados_libertad['y'].fuerza -= self._c * (-q_1 * self._L ** 2 / 12 - (
                    q_2 - q_1) * self._L ** 2 / 20) * 3.0 / (2.0 * self._L)
            self._fuerzas_i_rot[1, 0] -= self._c * (
                    -q_1 * self._L ** 2 / 12 - (q_2 - q_1) * self._L ** 2 / 20) * 3.0 / (
                                                 2.0 * self._L)
            self._nodo_i.grados_libertad['eje_z'].fuerza -= 0.5 * (
                    -q_1 * self._L ** 2 / 12 - (q_2 - q_1) * self._L ** 2 / 20)
            self._fuerzas_i_rot[2, 0] -= 0.5 * (-q_1 * self._L ** 2 / 12 - (q_2 - q_1) * self._L ** 2 / 20)
            self._nodo_j.grados_libertad['x'].fuerza -= -self._s * -(
                    -q_1 * self._L ** 2 / 12 - (q_2 - q_1) * self._L ** 2 / 20) * 3.0 / (2.0 * self._L)
            self._fuerzas_j_rot[0, 0] -= -self._s * -(
                    -q_1 * self._L ** 2 / 12 - (q_2 - q_1) * self._L ** 2 / 20) * 3.0 / (
                                                 2.0 * self._L)
            self._nodo_j.grados_libertad['y'].fuerza -= self._c * -(
                    -q_1 * self._L ** 2 / 12 - (q_2 - q_1) * self._L ** 2 / 20) * 3.0 / (2.0 * self._L)
            self._fuerzas_j_rot[1, 0] -= self._c * -(
                    -q_1 * self._L ** 2 / 12 - (q_2 - q_1) * self._L ** 2 / 20) * 3.0 / (
                                                 2.0 * self._L)
            self._nodo_j.grados_libertad['eje_z'].fuerza -= -q_1 * self._L ** 2 / 12 - (q_2 - q_1) * self._L ** 2 / 20
            self._fuerzas_j_rot[2, 0] -= -q_1 * self._L ** 2 / 12 - (q_2 - q_1) * self._L ** 2 / 20

    def agregar_carga_puntual(self, p: float, a: float = None):
        """Agrega una carga puntual perpendicular al elemento a una distancia 'a'.

        Calcula las fuerzas y momentos de empotramiento perfecto (fixed-end actions)
        y los suma a los vectores de fuerza nodal. Si el nodo final es una rótula,
        se aplica una corrección.

        Parameters
        ----------
        p : float
            Magnitud de la carga puntual. Un valor positivo indica que la carga
            actúa en la dirección positiva del eje 'y' local del elemento.
        a : float, optional
            Distancia desde el nodo inicial (i) a la que se aplica la carga.
            Si es None, la carga se aplica en el centro del elemento (L/2).
            Por defecto es None.
        """
        if a is None:
            a = self._L / 2

        # Fuerzas y momentos de empotramiento perfecto para carga puntual
        self._nodo_i.grados_libertad['x'].fuerza += -p * (self._L - a) ** 2 * (self._L + 2 * a) * self._s / self._L ** 3
        self._fuerzas_i[0, 0] += -p * (self._L - a) ** 2 * (self._L + 2 * a) * self._s / self._L ** 3
        self._nodo_i.grados_libertad['y'].fuerza += p * (self._L - a) ** 2 * (self._L + 2 * a) * self._c / self._L ** 3
        self._fuerzas_i[1, 0] += p * (self._L - a) ** 2 * (self._L + 2 * a) * self._c / self._L ** 3
        self._nodo_i.grados_libertad['eje_z'].fuerza += p * a * (self._L - a) ** 2 / self._L ** 2
        self._fuerzas_i[2, 0] += p * a * (self._L - a) ** 2 / self._L ** 2
        self._nodo_j.grados_libertad['x'].fuerza += -p * a ** 2 * (self._L + 2 * (self._L - a)) * self._s / self._L ** 3
        self._fuerzas_j[0, 0] += -p * a ** 2 * (self._L + 2 * (self._L - a)) * self._s / self._L ** 3
        self._nodo_j.grados_libertad['y'].fuerza += p * a ** 2 * (self._L + 2 * (self._L - a)) * self._c / self._L ** 3
        self._fuerzas_j[1, 0] += p * a ** 2 * (self._L + 2 * (self._L - a)) * self._c / self._L ** 3
        self._nodo_j.grados_libertad['eje_z'].fuerza += -p * (self._L - a) * a ** 2 / self._L ** 2
        self._fuerzas_j[2, 0] += -p * (self._L - a) * a ** 2 / self._L ** 2

        self._cargas['puntual'].append([p, a])
        self._cargas['puntual'].sort(key=lambda l: l[1])

        # Corrección por rótula en el nodo j
        if self._nodo_j.es_rotula:
            self._nodo_i.grados_libertad['x'].fuerza -= -self._s * (
                    -p * (self._L - a) * a ** 2 / self._L ** 2) * 3.0 / (
                                                                2.0 * self._L)
            self._fuerzas_i_rot[0, 0] -= -self._s * (-p * (self._L - a) * a ** 2 / self._L ** 2) * 3.0 / (2.0 * self._L)
            self._nodo_i.grados_libertad['y'].fuerza -= self._c * (-p * (self._L - a) * a ** 2 / self._L ** 2) * 3.0 / (
                    2.0 * self._L)
            self._fuerzas_i_rot[1, 0] -= self._c * (-p * (self._L - a) * a ** 2 / self._L ** 2) * 3.0 / (2.0 * self._L)
            self._nodo_i.grados_libertad['eje_z'].fuerza -= 0.5 * (-p * (self._L - a) * a ** 2 / self._L ** 2)
            self._fuerzas_i_rot[2, 0] -= 0.5 * (-p * (self._L - a) * a ** 2 / self._L ** 2)
            self._nodo_j.grados_libertad['x'].fuerza -= -self._s * -(
                    -p * (self._L - a) * a ** 2 / self._L ** 2) * 3.0 / (2.0 * self._L)
            self._fuerzas_j_rot[0, 0] -= -self._s * -(-p * (self._L - a) * a ** 2 / self._L ** 2) * 3.0 / (
                    2.0 * self._L)
            self._nodo_j.grados_libertad['y'].fuerza -= self._c * -(
                    -p * (self._L - a) * a ** 2 / self._L ** 2) * 3.0 / (
                                                                2.0 * self._L)
            self._fuerzas_j_rot[1, 0] -= self._c * -(-p * (self._L - a) * a ** 2 / self._L ** 2) * 3.0 / (2.0 * self._L)
            self._nodo_j.grados_libertad['eje_z'].fuerza -= -p * (self._L - a) * a ** 2 / self._L ** 2
            self._fuerzas_j_rot[2, 0] -= -p * (self._L - a) * a ** 2 / self._L ** 2

    def agregar_momento(self, m: float, a: float = None):
        m *= -1.0
        if a is None:
            a = self._L / 2
        self._nodo_i.grados_libertad['x'].fuerza += -6 * m * a * (self._L - a) * self._s / self._L ** 3
        self._fuerzas_i[0, 0] += -6 * m * a * (self._L - a) * self._s / self._L ** 3
        self._nodo_i.grados_libertad['y'].fuerza += 6 * m * a * (self._L - a) * self._c / self._L ** 3
        self._fuerzas_i[1, 0] += 6 * m * a * (self._L - a) * self._c / self._L ** 3
        self._nodo_i.grados_libertad['eje_z'].fuerza += m * (self._L - a) * (3 * a - self._L) / self._L ** 2
        self._fuerzas_i[2, 0] += m * (self._L - a) * (3 * a - self._L) / self._L ** 2
        self._nodo_j.grados_libertad['x'].fuerza += 6 * m * a * (self._L - a) * self._s / self._L ** 3
        self._fuerzas_j[0, 0] +=  6 * m * a * (self._L - a) * self._s / self._L ** 3
        self._nodo_j.grados_libertad['y'].fuerza += - 6 * m * a * (self._L - a) * self._c / self._L ** 3
        self._fuerzas_j[1, 0] += - 6 * m * a * (self._L - a) * self._c / self._L ** 3
        self._nodo_j.grados_libertad['eje_z'].fuerza += m * a * (2 * self._L - 3 * a) / self._L ** 2
        self._fuerzas_j[2, 0] += m * a * (2 * self._L - 3 * a) / self._L ** 2

        self._cargas['momento'].append([-m, a])
        self._cargas['momento'].sort(key=lambda l: l[1])

        if self._nodo_j.es_rotula:
            self._nodo_i.grados_libertad['x'].fuerza -= -self._s * (m * a * (2 * self._L - 3 * a) / self._L ** 2) * 3.0 / (
                                                                2.0 * self._L)
            self._fuerzas_i_rot[0, 0] -= -self._s * (m * a * (2 * self._L - 3 * a) / self._L ** 2) * 3.0 / (2.0 * self._L)
            self._nodo_i.grados_libertad['y'].fuerza -= self._c * (m * a * (2 * self._L - 3 * a) / self._L ** 2) * 3.0 / (
                    2.0 * self._L)
            self._fuerzas_i_rot[1, 0] -= self._c * (m * a * (2 * self._L - 3 * a) / self._L ** 2) * 3.0 / (2.0 * self._L)
            self._nodo_i.grados_libertad['eje_z'].fuerza -= 0.5 * (m * a * (2 * self._L - 3 * a) / self._L ** 2)
            self._fuerzas_i_rot[2, 0] -= 0.5 * (m * a * (2 * self._L - 3 * a) / self._L ** 2)
            self._nodo_j.grados_libertad['x'].fuerza -= -self._s * -(m * a * (2 * self._L - 3 * a) / self._L ** 2) * 3.0 / (2.0 * self._L)
            self._fuerzas_j_rot[0, 0] -= -self._s * -(m * a * (2 * self._L - 3 * a) / self._L ** 2) * 3.0 / (
                    2.0 * self._L)
            self._nodo_j.grados_libertad['y'].fuerza -= self._c * -(m * a * (2 * self._L - 3 * a) / self._L ** 2) * 3.0 / (
                                                                2.0 * self._L)
            self._fuerzas_j_rot[1, 0] -= self._c * -(m * a * (2 * self._L - 3 * a) / self._L ** 2) * 3.0 / (2.0 * self._L)
            self._nodo_j.grados_libertad['eje_z'].fuerza -= m * a * (2 * self._L - 3 * a) / self._L ** 2
            self._fuerzas_j_rot[2, 0] -= m * a * (2 * self._L - 3 * a) / self._L ** 2

    def agregar_carga_triangular_descendente(self, q: float):
        self._nodo_i.grados_libertad['x'].fuerza += -7 * q * self._L * self._s / 20
        self._fuerzas_i[0, 0] += -7 * q * self._L * self._s / 20
        self._nodo_i.grados_libertad['y'].fuerza += 7 * q * self._L * self._c / 20
        self._fuerzas_i[1, 0] += 7 * q * self._L * self._c / 20
        self._nodo_i.grados_libertad['eje_z'].fuerza += q * self._L ** 2 / 20
        self._fuerzas_i[2, 0] += q * self._L ** 2 / 20
        self._nodo_j.grados_libertad['x'].fuerza += -3 * q * self._L * self._s / 20
        self._fuerzas_j[0, 0] += -3 * q * self._L * self._s / 20
        self._nodo_j.grados_libertad['y'].fuerza += 3 * q * self._L * self._c / 20
        self._fuerzas_j[1, 0] += 3 * q * self._L * self._c / 20
        self._nodo_j.grados_libertad['eje_z'].fuerza += -q * self._L ** 2 / 30
        self._fuerzas_j[2, 0] += -q * self._L ** 2 / 30

        self._cargas['distribuida'].append([q, 0])

        if self._nodo_j.es_rotula:
            self._nodo_i.grados_libertad['x'].fuerza -= -self._s * (-q * self._L ** 2 / 30) * 3.0 / (2.0 * self._L)
            self._fuerzas_i_rot[0, 0] -= -self._s * (-q * self._L ** 2 / 30) * 3.0 / (2.0 * self._L)
            self._nodo_i.grados_libertad['y'].fuerza -= self._c * (-q * self._L ** 2 / 30) * 3.0 / (2.0 * self._L)
            self._fuerzas_i_rot[1, 0] -= self._c * (-q * self._L ** 2 / 30) * 3.0 / (2.0 * self._L)
            self._nodo_i.grados_libertad['eje_z'].fuerza -= 0.5 * (-q * self._L ** 2 / 30)
            self._fuerzas_i_rot[2, 0] -= 0.5 * (-q * self._L ** 2 / 30)
            self._nodo_j.grados_libertad['x'].fuerza -= -self._s * -(-q * self._L ** 2 / 30) * 3.0 / (2.0 * self._L)
            self._fuerzas_j_rot[0, 0] -= -self._s * -(-q * self._L ** 2 / 30) * 3.0 / (2.0 * self._L)
            self._nodo_j.grados_libertad['y'].fuerza -= self._c * -(-q * self._L ** 2 / 30) * 3.0 / (2.0 * self._L)
            self._fuerzas_j_rot[1, 0] -= self._c * -(-q * self._L ** 2 / 30) * 3.0 / (2.0 * self._L)
            self._nodo_j.grados_libertad['eje_z'].fuerza -= -q * self._L ** 2 / 30
            self._fuerzas_j_rot[2, 0] -= -q * self._L ** 2 / 30

    def agregar_carga_triangular_ascendente(self, q: float):
        self._nodo_i.grados_libertad['x'].fuerza += -3 * q * self._L * self._s / 20
        self._fuerzas_i[0, 0] += -3 * q * self._L * self._s / 20
        self._nodo_i.grados_libertad['y'].fuerza += 3 * q * self._L * self._c / 20
        self._fuerzas_i[1, 0] += 3 * q * self._L * self._c / 20
        self._nodo_i.grados_libertad['eje_z'].fuerza += q * self._L ** 2 / 30
        self._fuerzas_i[2, 0] += q * self._L ** 2 / 30
        self._nodo_j.grados_libertad['x'].fuerza += -7 * q * self._L * self._s / 20
        self._fuerzas_j[0, 0] += -7 * q * self._L * self._s / 20
        self._nodo_j.grados_libertad['y'].fuerza += 7 * q * self._L * self._c / 20
        self._fuerzas_j[1, 0] += 7 * q * self._L * self._c / 20
        self._nodo_j.grados_libertad['eje_z'].fuerza += - q * self._L ** 2 / 20
        self._fuerzas_j[2, 0] += - q * self._L ** 2 / 20

        self._cargas['distribuida'].append([0, q])

        if self._nodo_j.es_rotula:
            self._nodo_i.grados_libertad['x'].fuerza -= -self._s * (-q * self._L ** 2 / 20) * 3.0 / (2.0 * self._L)
            self._fuerzas_i_rot[0, 0] -= -self._s * (-q * self._L ** 2 / 20) * 3.0 / (2.0 * self._L)
            self._nodo_i.grados_libertad['y'].fuerza -= self._c * (-q * self._L ** 2 / 20) * 3.0 / (2.0 * self._L)
            self._fuerzas_i_rot[1, 0] -= self._c * (-q * self._L ** 2 / 20) * 3.0 / (2.0 * self._L)
            self._nodo_i.grados_libertad['eje_z'].fuerza -= 0.5 * (-q * self._L ** 2 / 20)
            self._fuerzas_i_rot[2, 0] -= 0.5 * (-q * self._L ** 2 / 20)
            self._nodo_j.grados_libertad['x'].fuerza -= -self._s * -(-q * self._L ** 2 / 20) * 3.0 / (2.0 * self._L)
            self._fuerzas_j_rot[0, 0] -= -self._s * -(-q * self._L ** 2 / 20) * 3.0 / (2.0 * self._L)
            self._nodo_j.grados_libertad['y'].fuerza -= self._c * -(-q * self._L ** 2 / 20) * 3.0 / (2.0 * self._L)
            self._fuerzas_j_rot[1, 0] -= self._c * -(-q * self._L ** 2 / 20) * 3.0 / (2.0 * self._L)
            self._nodo_j.grados_libertad['eje_z'].fuerza -= -q * self._L ** 2 / 20
            self._fuerzas_j_rot[2, 0] -= -q * self._L ** 2 / 20

    def agregar_carga_axial(self, p: float, a: float = None):
        if a is None:
            a = self._L / 2
        self._nodo_i.grados_libertad['x'].fuerza += p * self._c / 2
        self._fuerzas_i[0, 0] += p * self._c / 2
        self._nodo_i.grados_libertad['y'].fuerza += p * self._s / 2
        self._fuerzas_i[1, 0] += p * self._s / 2
        self._nodo_j.grados_libertad['x'].fuerza += p * self._c / 2
        self._fuerzas_j[0, 0] += p * self._c / 2
        self._nodo_j.grados_libertad['y'].fuerza += p * self._s / 2
        self._fuerzas_j[1, 0] += p * self._s / 2

        self._cargas['axial'].append([p, a])
        self._cargas['axial'].sort(key=lambda l: l[1])

    def _obtener_cargas(self) -> dict:
        return self._cargas

    def _calcular_coeficientes_diagramas(self):
        x_1 = 0.0
        x_2 = self._L
        q_1 = q_2 = 0
        for c in self._cargas['distribuida']:
            q_1 += c[0]
            q_2 += c[1]
        q = [0.0, 0.0]
        for c in self._cargas['puntual']:
            q[0] -= c[0] * (self._L - c[1]) ** 3 / 6
            q[1] -= c[0] * (self._L - c[1]) ** 2 / 2
        for m in self._cargas['momento']:
            q[0] -= -m[0] * (self._L - m[1]) ** 2 / 2
            q[1] -= -m[0] * (self._L - m[1])
        self._coef['A'] = -(q_2 - q_1) / (x_2 - x_1)
        self._coef['B'] = -q_1 - self._coef['A'] * x_1
        A = np.array([[x_1 ** 3 / 6.0, x_1 ** 2 / 2.0, x_1, 1.0], [x_1 ** 2 / 2.0, x_1, 1.0, 0.0],
                      [x_2 ** 3 / 6.0, x_2 ** 2 / 2.0, x_2, 1.0], [x_2 ** 2 / 2.0, x_2, 1.0, 0.0]], dtype=np.double)
        # Se rota los desplazamientos y eliminan los desplazamientos en x
        b = np.matmul(self._T, self._obtener_desplazamientos())[(1, 2, 4, 5), :] * self._E * self._I
        b += np.array([[self._coef['A'] * x_1 ** 5 / 120 + self._coef['B'] * x_1 ** 4 / 24],
                       [self._coef['A'] * x_1 ** 4 / 24 + self._coef['B'] * x_1 ** 3 / 6],
                       [self._coef['A'] * x_2 ** 5 / 120 + self._coef['B'] * x_2 ** 4 / 24 + q[0]],
                       [self._coef['A'] * x_2 ** 4 / 24 + self._coef['B'] * x_2 ** 3 / 6 + q[1]]])
        sol = Gauss(A, b, pivote_parcial=True)
        sol.x[abs(sol.x) < TOL_CERO] = 0.0
        self._coef['c_1'] = sol.x[0, 0]
        self._coef['c_2'] = sol.x[1, 0]
        self._coef['c_3'] = sol.x[2, 0]
        self._coef['c_4'] = sol.x[3, 0]

    def ecuacion_de_axial(self):
        x = sp.symbols('x')
        N = sp.Function('N')(x)
        x_1 = 0.0
        x_2 = self._L
        fuerzas_internas = (np.matmul(np.matmul(self._kp, self._T), self._obtener_desplazamientos()) -
                            np.matmul(self._T, self._obtener_fuerzas() + self._obtener_fuerzas_por_rotula()))
        ec = -fuerzas_internas[0, 0]
        if len(self._cargas['axial']) == 0:
            return sp.Eq(N, sp.Piecewise((ec, (x >= x_1) & (x <= x_2))))
        else:
            dis = [float(x_1)] + [float(x_1 + c[1]) for c in self._cargas['axial']] + [float(x_2)]
            car = [ec]
            for c in self._cargas['axial']:
                car.append(car[-1] - c[0])
            tramos = []
            for i in range(len(dis) - 1):
                tramos.append((car[i], (dis[i] <= x) & (x <= dis[i + 1])))
            return sp.Eq(N, sp.Piecewise(*tramos))

    def ecuacion_de_cortante(self):
        x = sp.symbols('x')
        V = sp.Function('V')(x)
        self._calcular_coeficientes_diagramas()
        x_1 = 0.0
        x_2 = self._L
        ec = -self._coef['A'] * x ** 2 / 2 - self._coef['B'] * x + self._coef['c_1']
        if len(self._cargas['puntual']) == 0:
            return sp.Eq(V, sp.Piecewise((ec, (x >= x_1) & (x <= x_2))))
        else:
            dis = [float(x_1)] + [float(x_1 + c[1]) for c in self._cargas['puntual']] + [float(x_2)]
            car = [ec]
            for c in self._cargas['puntual']:
                car.append(car[-1] + c[0])
            tramos = []
            for i in range(len(dis) - 1):
                tramos.append((car[i], (dis[i] <= x) & (x <= dis[i + 1])))
            return sp.Eq(V, sp.Piecewise(*tramos))

    def ecuacion_de_momento(self):
        x = sp.symbols('x')
        M = sp.Function('M')(x)
        self._calcular_coeficientes_diagramas()
        x_1 = 0.0
        x_2 = self._L
        ec = -self._coef['A'] * x ** 3 / 6 - self._coef['B'] * x ** 2 / 2 + self._coef['c_1'] * x + self._coef[
            'c_2']
        if len(self._cargas['puntual']) == 0 and len(self._cargas['momento']) == 0:
            return sp.Eq(M, sp.Piecewise((ec, (x >= x_1) & (x <= x_2))))
        else:
            cargas_ele = dict()
            for c in self._cargas['puntual']:
                cargas_ele[c[1]] = sp.expand(c[0] * (x - (x_1 + c[1])))
            for m in self._cargas['momento']:
                cargas_ele[m[1]] = sp.expand(cargas_ele.get(m[1], 0.0) - m[0])
            if len(cargas_ele) > 0:
                cargas_ele = dict(sorted(cargas_ele.items(), key=lambda item: item[0]))
            dis = [float(x_1)] + [float(x_1 + c) for c in cargas_ele.keys()] + [float(x_2)]
            #dis = [float(x_1)] + [float(x_1 + c[1]) for c in self._cargas['puntual']] + [float(x_2)]
            car = [ec]
            for c in cargas_ele.values():
                car.append(sp.expand(car[-1] + c))
            tramos = []
            for i in range(len(dis) - 1):
                tramos.append((car[i], (dis[i] <= x) & (x <= dis[i + 1])))
            return sp.Eq(M, sp.Piecewise(*tramos))

    def ecuacion_de_giro(self):
        x = sp.symbols('x')
        phi = sp.Function('phi')(x)
        self._calcular_coeficientes_diagramas()
        x_1 = 0.0
        x_2 = self._L
        ec = (-self._coef['A'] * x ** 4 / 24 - self._coef['B'] * x ** 3 / 6 + self._coef['c_1'] * x ** 2 / 2 +
              self._coef['c_2'] * x + self._coef['c_3']) / self._E / self._I
        if len(self._cargas['puntual']) == 0 and len(self._cargas['momento']) == 0:
            return sp.Eq(phi, sp.Piecewise((ec, (x >= x_1) & (x <= x_2))))
        else:
            cargas_ele = dict()
            for c in self._cargas['puntual']:
                cargas_ele[c[1]] = sp.expand((c[0] * (x - (x_1 + c[1])) ** 2 / 2) / self._E / self._I)
            for m in self._cargas['momento']:
                cargas_ele[m[1]] = sp.expand(cargas_ele.get(m[1], 0.0) - m[0] * (x - (x_1 + m[1])) / self._E / self._I)
            if len(cargas_ele) > 0:
                cargas_ele = dict(sorted(cargas_ele.items(), key=lambda item: item[0]))
            dis = [float(x_1)] + [float(x_1 + c) for c in cargas_ele.keys()] + [float(x_2)]
            # dis = [float(x_1)] + [float(x_1 + c[1]) for c in self._cargas['puntual']] + [float(x_2)]
            car = [ec]
            for c in cargas_ele.values():
                car.append(sp.expand(car[-1] + c))
            # for c in self._cargas['puntual']:
            #     car.append(sp.expand(car[-1] + (c[0] * (x - (x_1 + c[1])) ** 2 / 2) / self._E / self._I))
            tramos = []
            for i in range(len(dis) - 1):
                tramos.append((car[i], (dis[i] <= x) & (x <= dis[i + 1])))
            return sp.Eq(phi, sp.Piecewise(*tramos))

    def ecuacion_de_deflexion(self):
        x = sp.symbols('x')
        y = sp.Function('y')(x)
        self._calcular_coeficientes_diagramas()
        x_1 = 0.0
        x_2 = self._L
        ec = (-self._coef['A'] * x ** 5 / 120 - self._coef['B'] * x ** 4 / 24 + self._coef['c_1'] * x ** 3 / 6 +
              self._coef['c_2'] * x ** 2 / 2 + self._coef['c_3'] * x + self._coef['c_4']) / self._E / self._I
        x = sp.symbols('x')
        if len(self._cargas['puntual']) == 0 and len(self._cargas['momento']) == 0:
            return sp.Eq(y, sp.Piecewise((ec, (x >= x_1) & (x <= x_2))))
        else:
            cargas_ele = dict()
            for c in self._cargas['puntual']:
                cargas_ele[c[1]] = sp.expand((c[0] * (x - (x_1 + c[1])) ** 3 / 6) / self._E / self._I)
            for m in self._cargas['momento']:
                cargas_ele[m[1]] = sp.expand(
                    cargas_ele.get(m[1], 0.0) - (m[0] * (x - (x_1 + m[1])) ** 2 / 2) / self._E / self._I)
            if len(cargas_ele) > 0:
                cargas_ele = dict(sorted(cargas_ele.items(), key=lambda item: item[0]))
            dis = [float(x_1)] + [float(x_1 + c) for c in cargas_ele.keys()] + [float(x_2)]
            # dis = [float(x_1)] + [float(x_1 + c[1]) for c in self._cargas['puntual']] + [float(x_2)]
            car = [ec]
            for c in cargas_ele.values():
                car.append(sp.expand(car[-1] + c))

            # car = [ec]
            # for c in self._cargas['puntual']:
            #     car.append(sp.expand(car[-1] + (c[0] * (x - (x_1 + c[1])) ** 3 / 6) / self._E / self._I))
            tramos = []
            for i in range(len(dis) - 1):
                tramos.append((car[i], (dis[i] <= x) & (x <= dis[i + 1])))
            return sp.Eq(y, sp.Piecewise(*tramos))

    def _obtener_arrays_axiales(self, n_puntos: int = 100):
        x = sp.symbols('x')
        f = self.ecuacion_de_axial()
        x_1 = 0.0
        x_2 = self._L
        lista_x = np.linspace(x_1, x_2, n_puntos)
        ev = sp.lambdify(x, f.args[1], 'numpy')
        lista_n = ev(lista_x)
        lista_p = []
        for c in self._cargas['axial']:
            for fun in f.args[1].args:
                val, b = fun.subs(x, x_1 + c[1])
                if b:
                    lista_p.append((x_1 + c[1], val.evalf()))
        lista_p = [(x_1, ev(x_1))] + lista_p + [(x_2, ev(x_2))]
        lista_p = np.array(lista_p, dtype=float)
        lista_p[:, 1][abs(lista_p[:, 1]) < TOL_CERO] = 0.0
        return lista_x, lista_n, lista_p

    def _obtener_arrays_cortantes(self, n_puntos: int = 100):
        x = sp.symbols('x')
        f = self.ecuacion_de_cortante()
        x_1 = 0.0
        x_2 = self._L
        lista_x = np.linspace(x_1, x_2, n_puntos)
        ev = sp.lambdify(x, f.args[1], 'numpy')
        lista_v = ev(lista_x)
        der = sp.diff(f.args[1])
        p_crit = []
        for i in der.args:
            if sp.degree(i[0]) >= 1:
                p_crit += sp.real_roots(i[0], x)
        #lista_p = [[x_1 + c[1], ev(x_1 + c[1])] for c in self._cargas['puntual']]
        lista_p = []
        for c in self._cargas['puntual']:
            for fun in f.args[1].args:
                val, b = fun.subs(x, x_1 + c[1])
                if b:
                    lista_p.append((x_1 + c[1], val.evalf()))
        lista_p = [(x_1, ev(x_1))] + lista_p + [(x_2, ev(x_2))]
        for item in p_crit:
            lista_p.append((item, f.args[1].subs(x, item).evalf()))
        lista_p = np.array(lista_p, dtype=float)
        lista_p[:, 1][abs(lista_p[:, 1]) < TOL_CERO] = 0.0
        return lista_x, lista_v, lista_p

    def _obtener_arrays_momentos(self, n_puntos: int = 100):
        x = sp.symbols('x')
        f = self.ecuacion_de_momento()
        x_1 = 0.0
        x_2 = self._L
        lista_x = np.linspace(x_1, x_2, n_puntos)
        ev = sp.lambdify(x, f.args[1], 'numpy')
        lista_m = ev(lista_x)
        der = sp.diff(f.args[1])
        p_crit = []
        for i in der.args:
            if sp.degree(i[0]) >= 1:
                p_crit += sp.real_roots(i[0], x)
        lista_p = [(x_1 + c[1], ev(x_1 + c[1])) for c in self._cargas['puntual']]
        for m in self._cargas['momento']:
            for fun in f.args[1].args:
                val, b = fun.subs(x, x_1 + m[1])
                if b:
                    lista_p.append((x_1 + m[1], val.evalf()))
        # lista_p = []
        # for c in self._cargas['puntual']:
        #     for fun in f.args[1].args:
        #         val, b = fun.subs(x, x_1 + c[1])
        #         if b:
        #             lista_p.append((x_1 + c[1], val.evalf()))
        lista_p = [(x_1, ev(x_1))] + lista_p + [(x_2, ev(x_2))]
        for item in p_crit:
            lista_p.append((item, f.args[1].subs(x, item).evalf()))
        lista_p = np.array(lista_p, dtype=float)
        lista_p[:, 1][abs(lista_p[:, 1]) < TOL_CERO] = 0.0
        return lista_x, lista_m, lista_p

    def _obtener_arrays_angulos(self, n_puntos: int = 100):
        x = sp.symbols('x')
        f = self.ecuacion_de_giro()
        x_1 = 0.0
        x_2 = self._L
        lista_x = np.linspace(x_1, x_2, n_puntos)
        ev = sp.lambdify(x, f.args[1], 'numpy')
        lista_o = ev(lista_x)
        der = sp.diff(f.args[1])
        p_crit = []
        for i in der.args:
            if sp.degree(i[0]) >= 1:
                p_crit += sp.real_roots(i[0], x)
        lista_p = [[x_1 + c[1], ev(x_1 + c[1])] for c in self._cargas['puntual']]
        lista_p += [(x_1 + m[1], ev(x_1 + m[1])) for m in self._cargas['momento']]
        # lista_p = []
        # for c in self._cargas['puntual']:
        #     for fun in f.args[1].args:
        #         val, b = fun.subs(x, x_1 + c[1])
        #         if b:
        #             lista_p.append((x_1 + c[1], val.evalf()))
        lista_p = [(x_1, ev(x_1))] + lista_p + [(x_2, ev(x_2))]
        for item in p_crit:
            lista_p.append((item, f.args[1].subs(x, item).evalf()))
        lista_p = np.array(lista_p, dtype=float)
        lista_p[:, 1][abs(lista_p[:, 1]) < TOL_CERO] = 0.0
        return lista_x, lista_o, lista_p

    def _obtener_arrays_deflexion(self, n_puntos: int = 100):
        x = sp.symbols('x')
        f = self.ecuacion_de_deflexion()
        x_1 = 0.0
        x_2 = self._L
        lista_x = np.linspace(x_1, x_2, n_puntos)
        ev = sp.lambdify(x, f.args[1], 'numpy')
        lista_y = ev(lista_x)
        der = sp.diff(f.args[1])
        p_crit = []
        for i in der.args:
            if sp.degree(i[0]) >= 1:
                p_crit += sp.real_roots(i[0], x)
        lista_p = [[x_1 + c[1], ev(x_1 + c[1])] for c in self._cargas['puntual']]
        lista_p += [(x_1 + m[1], ev(x_1 + m[1])) for m in self._cargas['momento']]
        # lista_p = []
        # for c in self._cargas['puntual']:
        #     for fun in f.args[1].args:
        #         val, b = fun.subs(x, x_1 + c[1])
        #         if b:
        #             lista_p.append((x_1 + c[1], val.evalf()))
        lista_p = [(x_1, ev(x_1))] + lista_p + [(x_2, ev(x_2))]
        for item in p_crit:
            lista_p.append((item, f.args[1].subs(x, item).evalf()))
        lista_p = np.array(lista_p, dtype=float)
        lista_p[:, 1][abs(lista_p[:, 1]) < TOL_CERO] = 0.0
        return lista_x, lista_y, lista_p

    def diagrama_de_axial(self, n_puntos: int = 100, local: bool = True):
        l_x, l_y, l_z = self._obtener_arrays_axiales(n_puntos)
        fig, ax = plt.subplots()
        if local:
            plt.fill_between(l_x, l_y, color='teal', lw=0.5, alpha=0.5)
            for i in l_z:
                pos_y = 5  # offset escritura
                val_x, val_y = i
                if val_y < 0:
                    pos_y = -5
                plt.annotate(f'${float(val_y):.4G}$', (val_x, val_y), c='black',
                             textcoords="offset points", xytext=(0, pos_y), va='center', ha='center', fontsize=8)
        else:
            ax.add_patch(patches.PathPatch(self.__obtener_path_elemento(self._L / 20), edgecolor='royalblue',
                                           facecolor='lightsteelblue', lw=0.2))
            max_val = max(abs(l_y))
            if abs(max_val) < 1E-12:
                factor = 0.0
            else:
                factor = 0.5 * self._L / max_val
            x = self._nodo_i.punto[0] + l_x * self._c
            y = self._nodo_i.punto[1] + l_x * self._s
            ly_prima = l_y * factor
            f_x = x - ly_prima * self._s
            f_y = y + ly_prima * self._c
            plt.plot(f_x, f_y, c='g', lw=0.1)
            for i in range(len(x)):
                plt.plot([f_x[i], x[i]], [f_y[i], y[i]], c='g', lw=0.1)
            for i in l_z:
                ang = np.degrees(np.arccos(self._c)) * np.sign(self._s)
                pos_y = 5  # offset escritura
                val_x, val_y = i
                if val_y < 0:
                    pos_y = -5
                x = self._nodo_i.punto[0] + val_x * self._c - val_y * factor * self._s
                y = self._nodo_i.punto[1] + val_x * self._s + val_y * factor * self._c
                plt.annotate(f'${float(val_y):.4G}$', (x, y), c='black',
                             textcoords="offset points", xytext=(-pos_y * self._s, pos_y * self._c), va='center',
                             ha='center', fontsize=8,
                             rotation=ang, rotation_mode='anchor')
            ax.axis('equal')
        plt.grid()
        plt.title('Diagrama de axiales')
        plt.xlabel('$x$')
        plt.ylabel('$N$')
        plt.show()

    def diagrama_de_cortante(self, n_puntos: int = 100, local: bool = True):
        l_x, l_y, l_z = self._obtener_arrays_cortantes(n_puntos)
        fig, ax = plt.subplots()
        if local:
            plt.fill_between(l_x, l_y, color='teal', lw=0.5, alpha=0.5)
            for i in l_z:
                pos_y = 5  # offset escritura
                val_x, val_y = i
                if val_y < 0:
                    pos_y = -5
                plt.annotate(f'${float(val_y):.4G}$', (val_x, val_y), c='black',
                             textcoords="offset points", xytext=(0, pos_y), va='center', ha='center', fontsize=8)
        else:
            ax.add_patch(patches.PathPatch(self.__obtener_path_elemento(self._L / 20), edgecolor='royalblue',
                                           facecolor='lightsteelblue', lw=0.2))
            max_val = max(abs(l_y))
            if abs(max_val) < 1E-12:
                factor = 0.0
            else:
                factor = 0.5 * self._L / max_val
            x = self._nodo_i.punto[0] + l_x * self._c
            y = self._nodo_i.punto[1] + l_x * self._s
            ly_prima = l_y * factor
            f_x = x - ly_prima * self._s
            f_y = y + ly_prima * self._c
            plt.plot(f_x, f_y, c='r', lw=0.1)
            for i in range(len(x)):
                plt.plot([f_x[i], x[i]], [f_y[i], y[i]], c='r', lw=0.1)
            for i in l_z:
                ang = np.degrees(np.arccos(self._c)) * np.sign(self._s)
                pos_y = 5  # offset escritura
                val_x, val_y = i
                if val_y < 0:
                    pos_y = -5
                x = self._nodo_i.punto[0] + val_x * self._c - val_y * factor * self._s
                y = self._nodo_i.punto[1] + val_x * self._s + val_y * factor * self._c
                plt.annotate(f'${float(val_y):.4G}$', (x, y), c='black',
                             textcoords="offset points", xytext=(-pos_y * self._s, pos_y * self._c), va='center',
                             ha='center', fontsize=8,
                             rotation=ang, rotation_mode='anchor')
            ax.axis('equal')
        plt.grid()
        plt.title('Diagrama de cortantes')
        plt.xlabel('$x$')
        plt.ylabel('$V$')
        plt.show()

    def diagrama_de_momento(self, n_puntos: int = 100, local: bool = True):
        l_x, l_y, l_z = self._obtener_arrays_momentos(n_puntos)
        fig, ax = plt.subplots()
        if local:
            plt.fill_between(l_x, l_y, color='teal', lw=0.5, alpha=0.5)
            for i in l_z:
                pos_y = 5  # offset escritura
                val_x, val_y = i
                if val_y < 0:
                    pos_y = -5
                plt.annotate(f'${float(val_y):.4G}$', (val_x, val_y), c='black',
                             textcoords="offset points", xytext=(0, pos_y), va='center', ha='center', fontsize=8)
        else:
            ax.add_patch(patches.PathPatch(self.__obtener_path_elemento(self._L / 20), edgecolor='royalblue',
                                           facecolor='lightsteelblue', lw=0.2))
            max_val = max(abs(l_y))
            if abs(max_val) < 1E-12:
                factor = 0.0
            else:
                factor = 0.5 * self._L / max_val
            x = self._nodo_i.punto[0] + l_x * self._c
            y = self._nodo_i.punto[1] + l_x * self._s
            ly_prima = l_y * factor
            f_x = x - ly_prima * self._s
            f_y = y + ly_prima * self._c
            plt.plot(f_x, f_y, c='r', lw=0.1)
            for i in range(len(x)):
                plt.plot([f_x[i], x[i]], [f_y[i], y[i]], c='r', lw=0.1)
            for i in l_z:
                ang = np.degrees(np.arccos(self._c)) * np.sign(self._s)
                pos_y = 5  # offset escritura
                val_x, val_y = i
                if val_y < 0:
                    pos_y = -5
                x = self._nodo_i.punto[0] + val_x * self._c - val_y * factor * self._s
                y = self._nodo_i.punto[1] + val_x * self._s + val_y * factor * self._c
                plt.annotate(f'${float(val_y):.4G}$', (x, y), c='black',
                             textcoords="offset points", xytext=(-pos_y * self._s, pos_y * self._c), va='center',
                             ha='center', fontsize=8,
                             rotation=ang, rotation_mode='anchor')
            ax.axis('equal')
        plt.grid()
        plt.title('Diagrama de momentos')
        plt.xlabel('$x$')
        plt.ylabel('$M$')
        plt.show()

    def diagrama_de_giro(self, n_puntos: int = 100, local: bool = True):
        l_x, l_y, l_z = self._obtener_arrays_angulos(n_puntos)
        fig, ax = plt.subplots()
        if local:
            plt.fill_between(l_x, l_y, color='teal', lw=0.5, alpha=0.5)
            for i in l_z:
                pos_y = 5  # offset escritura
                val_x, val_y = i
                if val_y < 0:
                    pos_y = -5
                plt.annotate(f'${float(val_y):.4G}$', (val_x, val_y), c='black',
                             textcoords="offset points", xytext=(0, pos_y), va='center', ha='center', fontsize=8)
        else:
            ax.add_patch(patches.PathPatch(self.__obtener_path_elemento(self._L / 20), edgecolor='royalblue',
                                           facecolor='lightsteelblue', lw=0.2))
            max_val = max(abs(l_y))
            if abs(max_val) < 1E-12:
                factor = 0.0
            else:
                factor = 0.5 * self._L / max_val
            x = self._nodo_i.punto[0] + l_x * self._c
            y = self._nodo_i.punto[1] + l_x * self._s
            ly_prima = l_y * factor
            f_x = x - ly_prima * self._s
            f_y = y + ly_prima * self._c
            plt.plot(f_x, f_y, c='r', lw=0.1)
            for i in range(len(x)):
                plt.plot([f_x[i], x[i]], [f_y[i], y[i]], c='r', lw=0.1)
            for i in l_z:
                ang = np.degrees(np.arccos(self._c)) * np.sign(self._s)
                pos_y = 5  # offset escritura
                val_x, val_y = i
                if val_y < 0:
                    pos_y = -5
                x = self._nodo_i.punto[0] + val_x * self._c - val_y * factor * self._s
                y = self._nodo_i.punto[1] + val_x * self._s + val_y * factor * self._c
                plt.annotate(f'${float(val_y):.4G}$', (x, y), c='black',
                             textcoords="offset points", xytext=(-pos_y * self._s, pos_y * self._c), va='center',
                             ha='center', fontsize=8,
                             rotation=ang, rotation_mode='anchor')
            ax.axis('equal')
        plt.grid()
        plt.title('Diagrama de giro')
        plt.xlabel('$x$')
        plt.ylabel(r'$\phi$')
        plt.show()

    def diagrama_de_deflexion(self, n_puntos: int = 100, local: bool = True):
        l_x, l_y, l_z = self._obtener_arrays_deflexion(n_puntos)
        fig, ax = plt.subplots()
        if local:
            plt.fill_between(l_x, l_y, color='teal', lw=0.5, alpha=0.5)
            for i in l_z:
                pos_y = 5  # offset escritura
                val_x, val_y = i
                if val_y < 0:
                    pos_y = -5
                plt.annotate(f'${float(val_y):.4G}$', (val_x, val_y), c='black',
                             textcoords="offset points", xytext=(0, pos_y), va='center', ha='center', fontsize=8)
        else:
            ax.add_patch(patches.PathPatch(self.__obtener_path_elemento(self._L / 20), edgecolor='royalblue',
                                           facecolor='lightsteelblue', lw=0.2))
            max_val = max(abs(l_y))
            if abs(max_val) < 1E-12:
                factor = 0.0
            else:
                factor = 0.5 * self._L / max_val
            x = self._nodo_i.punto[0] + l_x * self._c
            y = self._nodo_i.punto[1] + l_x * self._s
            ly_prima = l_y * factor
            f_x = x - ly_prima * self._s
            f_y = y + ly_prima * self._c
            plt.plot(f_x, f_y, c='r', lw=0.1)
            for i in range(len(x)):
                plt.plot([f_x[i], x[i]], [f_y[i], y[i]], c='r', lw=0.1)
            for i in l_z:
                ang = np.degrees(np.arccos(self._c)) * np.sign(self._s)
                pos_y = 5  # offset escritura
                val_x, val_y = i
                if val_y < 0:
                    pos_y = -5
                x = self._nodo_i.punto[0] + val_x * self._c - val_y * factor * self._s
                y = self._nodo_i.punto[1] + val_x * self._s + val_y * factor * self._c
                plt.annotate(f'${float(val_y):.4G}$', (x, y), c='black',
                             textcoords="offset points", xytext=(-pos_y * self._s, pos_y * self._c), va='center',
                             ha='center', fontsize=8,
                             rotation=ang, rotation_mode='anchor')
            ax.axis('equal')
        plt.grid()
        plt.title('Diagrama de deflexión')
        plt.xlabel('$x$')
        plt.ylabel('$y$')
        plt.show()

    def fuerzas_internas(self):
        fuerzas_internas = np.matmul(self._k.k,
                                     self._obtener_desplazamientos()) - self._obtener_fuerzas() - self._obtener_fuerzas_por_rotula()
        fuerzas_internas[abs(fuerzas_internas) < TOL_CERO] = 0.0
        if es_notebook():
            indice = ['$' + label + '$' for label in self._obtener_etiquetas_fuerzas()]
            return tabulate({'Fuerzas internas': fuerzas_internas}, headers='keys', showindex=indice, tablefmt='html')
        else:
            print(fuerzas_internas)

    def _obtener_etiquetas_fuerzas(self, reducida: bool = False):
        def ajustar_cadena(s: str) -> str:
            if s == 'eje_z':
                return ''
            else:
                return s

        etq = [nodo.grados_libertad[
                   gl].label_fuerza + '^{(' + self.nombre + ')}_{' + nodo.nombre + ajustar_cadena(gl) + '}'
               for nodo in [self._nodo_i, self._nodo_j] for gl in self._k.grados if
               not reducida or nodo.grados_libertad[gl].valor]
        return etq

    def _obtener_etiquetas_fuerzas_local(self, reducida: bool = False):
        def ajustar_cadena(s: str) -> str:
            if s == 'eje_z':
                return ''
            else:
                return s

        etq = [nodo.grados_libertad[
                   gl].label_fuerza + r'^{\prime(' + self.nombre + ')}_{' + nodo.nombre + ajustar_cadena(gl) + '}'
               for nodo in [self._nodo_i, self._nodo_j] for gl in self._k.grados if
               not reducida or nodo.grados_libertad[gl].valor]
        return etq

    def fuerzas_internas_local(self):
        fuerzas_internas = (np.matmul(np.matmul(self._kp, self._T), self._obtener_desplazamientos()) -
                            np.matmul(self._T, self._obtener_fuerzas() + self._obtener_fuerzas_por_rotula()))
        fuerzas_internas[abs(fuerzas_internas) < TOL_CERO] = 0.0
        if es_notebook():
            indice = ['$' + label + '$' for label in self._obtener_etiquetas_fuerzas_local()]
            return tabulate({'Fuerzas internas local': fuerzas_internas}, headers='keys', showindex=indice,
                            tablefmt='html')
        else:
            print(fuerzas_internas)

    def _obtener_desplazamientos(self) -> np.ndarray:
        desplazamiento = None
        for item in self._k.lista_nodos:
            d_i = [d.desplazamiento for d in item.grados_libertad.values() if d.gl in self._k.grados]
            if desplazamiento is None:
                desplazamiento = d_i
            else:
                desplazamiento = np.hstack((desplazamiento, d_i))
        b = np.array(desplazamiento).reshape(-1, 1)
        # Si hay una rotula en el nodo derecho, se ajusta el ángulo
        if self._nodo_j.es_rotula:
            b[5] = (self._fuerzas_j[2] * self._L / self._E - np.matmul(np.array(
                [[- 6 * self._I * self._s / self._L, 6 * self._I * self._c / self._L, 2 * self._I,
                  6 * self._I * self._s / self._L, -6 * self._I * self._c / self._L]]), b[0:5, ])) / 4 / self._I
        return b

    def mostrar_sistema(self, reducida: bool = False):
        """Muestra el sistema de ecuaciones del elemento en formato matricial."""
        vec_d = np.array(self._k.obtener_etiquetas_desplazamientos(reducida), dtype=object).reshape(-1, 1)
        vec_f = np.array(self._obtener_etiquetas_fuerzas(reducida), dtype=object).reshape(-1, 1)
        if es_notebook():
            texto_latex = r'\left\{\begin{array}{c}'
            texto_latex += _generar_matrix(vec_f,
                                           '{:}') + r'\end{array}\right\}_{\{f\}}=\left[\begin{array}{' + 'c' * \
                           self._k.obtener_matriz(reducida).shape[1] + '}'
            texto_latex += _generar_matrix(self._k.obtener_matriz(reducida),
                                           '{:G}') + r'\end{array}\right]_{[k]}\cdot\left\{\begin{array}{c}'
            texto_latex += _generar_matrix(vec_d, '{:}') + r'\end{array}\right\}_{\{d\}}-\left\{\begin{array}{c}'
            texto_latex += _generar_matrix(self._obtener_fuerzas(reducida),
                                           '{:}') + r'\end{array}\right\}_{\{fo\}}'
            display(Math(texto_latex))
        else:
            return np.array2string(self._k.obtener_matriz(reducida),
                                   formatter={'float_kind': lambda x: '{:}'.format(x)})

    def get_matriz_rigidez_local(self):
        return self._kp

    def get_matriz_T(self):
        return self._T

    def __obtener_path_elemento(self, espesor: np.double) -> Path:
        t = espesor
        x_1, y_1, Z_1 = self._nodo_i.punto
        x_2, y_2, Z_2 = self._nodo_j.punto
        c = self._c
        s = self._s
        vertices = np.array([(x_1 + 0.5 * t * s, y_1 - 0.5 * t * c), (x_2 + 0.5 * t * s, y_2 - 0.5 * t * c),
                             (x_2 - 0.5 * t * s, y_2 + 0.5 * t * c), (x_1 - 0.5 * t * s, y_1 + 0.5 * t * c),
                             (x_1 + 0.5 * t * s, y_1 - 0.5 * t * c)])
        codes = [Path.MOVETO, Path.LINETO, Path.LINETO, Path.LINETO, Path.CLOSEPOLY]
        return Path(vertices, codes)

    def diagrama_fuerzas_internas(self):
        def dibujar_momento(p: tuple[float, float], r: float, teta_1: float = 0, teta_2: float = 90, ejes=None,
                            **kwargs):
            c_x, c_y = p
            r /= 72
            alfa = 0
            teta_flecha = teta_2
            if teta_2 < teta_1:
                teta_1, teta_2 = (teta_2, teta_1)
                alfa = 180
                teta_flecha = teta_1
            trans = (fig.dpi_scale_trans + transforms.ScaledTranslation(c_x, c_y, ejes.transData))
            arc = patches.Arc((0, 0), r, r, angle=0, theta1=teta_1, theta2=teta_2, transform=trans
                              , capstyle='round', linestyle='-', **kwargs)
            ejes.add_patch(arc)
            p_x = (r / 2) * np.cos(np.radians(teta_flecha))
            p_y = (r / 2) * np.sin(np.radians(teta_flecha))
            ejes.add_patch(
                patches.RegularPolygon(xy=(p_x, p_y), numVertices=3, radius=r / 20,
                                       orientation=np.radians(teta_flecha + alfa), transform=trans, **kwargs))

        fuerzas_internas = np.matmul(self._k.k,
                                     self._obtener_desplazamientos()) - self._obtener_fuerzas() - self._obtener_fuerzas_por_rotula()
        fuerzas_internas[abs(fuerzas_internas) < TOL_CERO] = 0.0
        etq = self._obtener_etiquetas_fuerzas()
        fig, ax = plt.subplots()
        ax.add_patch(patches.PathPatch(self.__obtener_path_elemento(self._L / 20), edgecolor='royalblue',
                                       facecolor='lightsteelblue', lw=0.2))
        fuerza = fuerzas_internas[0, 0]
        if fuerza <= 0.0:
            pos = -50
            ali = 'right'
            sentido = '<-'
        else:
            pos = -50
            ali = 'right'
            sentido = '->'
        ax.annotate('$' + etq[0] + '=' + '{:G}$'.format(abs(fuerza)),
                    xy=(self._nodo_i.punto[0:2]), xycoords='data',
                    xytext=(pos, 0), textcoords='offset points', va='center', ha=ali, size=12,
                    arrowprops=dict(arrowstyle=sentido, connectionstyle="arc3"))
        fuerza = fuerzas_internas[1, 0]
        if fuerza <= 0.0:
            pos = -50
            ali = 'top'
            sentido = '<-'
        else:
            pos = -50
            ali = 'top'
            sentido = '->'
        ax.annotate('$' + etq[1] + '=' + '{:G}$'.format(abs(fuerza)),
                    xy=(self._nodo_i.punto[0:2]), xycoords='data',
                    xytext=(0, pos), textcoords='offset points', va=ali, ha='center', size=12,
                    arrowprops=dict(arrowstyle=sentido, connectionstyle="arc3"))
        fuerza = fuerzas_internas[2, 0]
        if fuerza <= 0.0:
            dibujar_momento((self._nodo_i.punto[0], self._nodo_i.punto[1]), 50, 255, 30, ejes=ax, color='b', lw=2)
            pos = -20
            ali = 'right'
        else:
            dibujar_momento((self._nodo_i.punto[0], self._nodo_i.punto[1]), 50, -75, 150, ejes=ax, color='b', lw=2)
            pos = 20
            ali = 'left'
        ax.annotate('$' + etq[2] + '=' + '{:G}$'.format(abs(fuerza)),
                    xy=(self._nodo_i.punto[0:2]), xycoords='data',
                    xytext=(pos, 20), textcoords='offset points', va='bottom', ha=ali, size=12)
        fuerza = fuerzas_internas[3, 0]
        if fuerza <= 0.0:
            pos = 50
            ali = 'left'
            sentido = '->'
        else:
            pos = 50
            ali = 'left'
            sentido = '<-'
        ax.annotate('$' + etq[3] + '=' + '{:G}$'.format(abs(fuerza)),
                    xy=(self._nodo_j.punto[0:2]), xycoords='data',
                    xytext=(pos, 0), textcoords='offset points', va='center', ha=ali, size=12,
                    arrowprops=dict(arrowstyle=sentido, connectionstyle="arc3"))
        fuerza = fuerzas_internas[4, 0]
        if fuerza <= 0.0:
            pos = 50
            ali = 'bottom'
            sentido = '->'
        else:
            pos = 50
            ali = 'bottom'
            sentido = '<-'
        ax.annotate('$' + etq[4] + '=' + '{:G}$'.format(abs(fuerza)),
                    xy=(self._nodo_j.punto[0:2]), xycoords='data',
                    xytext=(0, pos), textcoords='offset points', va=ali, ha='center', size=12,
                    arrowprops=dict(arrowstyle=sentido, connectionstyle="arc3"))
        fuerza = fuerzas_internas[5, 0]
        if fuerza <= 0.0:
            dibujar_momento((self._nodo_j.punto[0], self._nodo_j.punto[1]), 50, 255, 30, ejes=ax, color='b', lw=2)
            pos = -20
            ali = 'right'
        else:
            dibujar_momento((self._nodo_j.punto[0], self._nodo_j.punto[1]), 50, -75, 150, ejes=ax, color='b', lw=2)
            pos = 20
            ali = 'left'
        ax.annotate('$' + etq[5] + '=' + '{:G}$'.format(abs(fuerza)),
                    xy=(self._nodo_j.punto[0:2]), xycoords='data',
                    xytext=(pos, 20), textcoords='offset points', va='bottom', ha=ali, size=12)
        ax.scatter([self._nodo_i.punto[0], self._nodo_j.punto[0]], [self._nodo_i.punto[1], self._nodo_j.punto[1]],
                   c='navy', marker='o')

        ax.axis('equal')
        ax.axis('off')
        ax.set_xmargin(0.15)
        ax.set_ymargin(0.15)
        plt.show()

    def diagrama_fuerzas_internas_local(self):
        def dibujar_momento(p: tuple[float, float], r: float, teta_1: float = 0, teta_2: float = 90, ejes=None,
                            **kwargs):
            c_x, c_y = p
            r /= 72
            alfa = 0
            teta_flecha = teta_2
            if teta_2 < teta_1:
                teta_1, teta_2 = (teta_2, teta_1)
                alfa = 180
                teta_flecha = teta_1
            trans = (fig.dpi_scale_trans + transforms.ScaledTranslation(c_x, c_y, ejes.transData))
            arc = patches.Arc((0, 0), r, r, angle=0, theta1=teta_1, theta2=teta_2, transform=trans
                              , capstyle='round', linestyle='-', **kwargs)
            ejes.add_patch(arc)
            p_x = (r / 2) * np.cos(np.radians(teta_flecha))
            p_y = (r / 2) * np.sin(np.radians(teta_flecha))
            ejes.add_patch(
                patches.RegularPolygon(xy=(p_x, p_y), numVertices=3, radius=r / 20,
                                       orientation=np.radians(teta_flecha + alfa), transform=trans, **kwargs))

        fuerzas_internas = (np.matmul(np.matmul(self._kp, self._T), self._obtener_desplazamientos()) -
                            np.matmul(self._T, self._obtener_fuerzas() + self._obtener_fuerzas_por_rotula()))
        fuerzas_internas[abs(fuerzas_internas) < TOL_CERO] = 0.0
        etq = self._obtener_etiquetas_fuerzas_local()
        fig, ax = plt.subplots()
        ax.add_patch(patches.PathPatch(self.__obtener_path_elemento(self._L / 20), edgecolor='royalblue',
                                       facecolor='lightsteelblue', lw=0.2))
        ang = np.degrees(np.arccos(self.get_coseno())) * np.sign(self.get_seno())
        fuerza = fuerzas_internas[0, 0]
        if fuerza <= 0.0:
            pos_x = -50 * self._c
            pos_y = -50 * self._s
            ali_v = 'center'
            ali_h = 'right'
            sentido = '<-'
        else:
            pos_x = -50 * self._c
            pos_y = -50 * self._s
            ali_v = 'center'
            ali_h = 'right'
            sentido = '->'
        ax.annotate('',
                    xy=(self._nodo_i.punto[0:2]), xycoords='data',
                    xytext=(pos_x, pos_y), textcoords='offset points', va=ali_v, ha=ali_h, size=12,
                    arrowprops=dict(arrowstyle=sentido, connectionstyle="arc3"))
        ax.annotate('$' + etq[0] + '=' + '{:G}$'.format(abs(fuerza)),
                    xy=(self._nodo_i.punto[0:2]), xycoords='data',
                    xytext=(pos_x, pos_y), textcoords='offset points', va=ali_v, ha=ali_h, size=12, rotation=ang,
                    rotation_mode='anchor')

        fuerza = fuerzas_internas[1, 0]
        if fuerza <= 0.0:
            pos_x = 50 * self._s
            pos_y = -50 * self._c
            ali_v = 'top'
            ali_h = 'center'
            sentido = '<-'
        else:
            pos_x = 50 * self._s
            pos_y = -50 * self._c
            ali_v = 'top'
            ali_h = 'center'
            sentido = '->'
        ax.annotate('',
                    xy=(self._nodo_i.punto[0:2]), xycoords='data',
                    xytext=(pos_x, pos_y), textcoords='offset points', va=ali_v, ha=ali_h, size=12,
                    arrowprops=dict(arrowstyle=sentido, connectionstyle="arc3"))
        ax.annotate('$' + etq[1] + '=' + '{:G}$'.format(abs(fuerza)),
                    xy=(self._nodo_i.punto[0:2]), xycoords='data',
                    xytext=(pos_x, pos_y), textcoords='offset points', va=ali_v, ha=ali_h, size=12, rotation=ang,
                    rotation_mode='anchor')
        fuerza = fuerzas_internas[2, 0]
        if fuerza <= 0.0:
            dibujar_momento((self._nodo_i.punto[0], self._nodo_i.punto[1]), 50, 255, 30, ejes=ax, color='b', lw=2)
            pos = -20
            ali = 'right'
        else:
            dibujar_momento((self._nodo_i.punto[0], self._nodo_i.punto[1]), 50, -75, 150, ejes=ax, color='b', lw=2)
            pos = 20
            ali = 'left'
        ax.annotate('$' + etq[2] + '=' + '{:G}$'.format(abs(fuerza)),
                    xy=(self._nodo_i.punto[0:2]), xycoords='data',
                    xytext=(pos, 20), textcoords='offset points', va='bottom', ha=ali, size=12, rotation=ang,
                    rotation_mode='anchor')

        fuerza = fuerzas_internas[3, 0]
        if fuerza <= 0.0:
            pos_x = 50 * self._c
            pos_y = 50 * self._s
            ali_v = 'center'
            ali_h = 'left'
            sentido = '->'
        else:
            pos_x = 50 * self._c
            pos_y = 50 * self._s
            ali_v = 'center'
            ali_h = 'left'
            sentido = '<-'
        ax.annotate('',
                    xy=(self._nodo_j.punto[0:2]), xycoords='data',
                    xytext=(pos_x, pos_y), textcoords='offset points', va=ali_v, ha=ali_h, size=12,
                    arrowprops=dict(arrowstyle=sentido, connectionstyle="arc3"))
        ax.annotate('$' + etq[3] + '=' + '{:G}$'.format(abs(fuerza)),
                    xy=(self._nodo_j.punto[0:2]), xycoords='data',
                    xytext=(pos_x, pos_y), textcoords='offset points', va=ali_v, ha=ali_h, size=12, rotation=ang,
                    rotation_mode='anchor')

        fuerza = fuerzas_internas[4, 0]
        if fuerza <= 0.0:
            pos_x = -50 * self._s
            pos_y = 50 * self._c
            ali_v = 'bottom'
            ali_h = 'center'
            sentido = '->'
        else:
            pos_x = -50 * self._s
            pos_y = 50 * self._c
            ali_v = 'bottom'
            ali_h = 'center'
            sentido = '<-'
        ax.annotate('',
                    xy=(self._nodo_j.punto[0:2]), xycoords='data',
                    xytext=(pos_x, pos_y), textcoords='offset points', va=ali_v, ha=ali_h, size=12,
                    arrowprops=dict(arrowstyle=sentido, connectionstyle="arc3"))
        ax.annotate('$' + etq[4] + '=' + '{:G}$'.format(abs(fuerza)),
                    xy=(self._nodo_j.punto[0:2]), xycoords='data',
                    xytext=(pos_x, pos_y), textcoords='offset points', va=ali_v, ha=ali_h, size=12, rotation=ang,
                    rotation_mode='anchor')
        fuerza = fuerzas_internas[5, 0]
        if fuerza <= 0.0:
            dibujar_momento((self._nodo_j.punto[0], self._nodo_j.punto[1]), 50, 255, 30, ejes=ax, color='b', lw=2)
            pos = -20
            ali = 'right'
        else:
            dibujar_momento((self._nodo_j.punto[0], self._nodo_j.punto[1]), 50, -75, 150, ejes=ax, color='b', lw=2)
            pos = 20
            ali = 'left'
        ax.annotate('$' + etq[5] + '=' + '{:G}$'.format(abs(fuerza)),
                    xy=(self._nodo_j.punto[0:2]), xycoords='data',
                    xytext=(pos, 20), textcoords='offset points', va='bottom', ha=ali, size=12, rotation=ang,
                    rotation_mode='anchor')
        plt.scatter([self._nodo_i.punto[0], self._nodo_j.punto[0]], [self._nodo_i.punto[1], self._nodo_j.punto[1]],
                    c='navy',
                    marker='o')

        ax.axis('equal')
        ax.axis('off')
        ax.set_xmargin(0.15)
        ax.set_ymargin(0.15)
        plt.show()



from mnspy.integrales import Integral
import numpy as np
import matplotlib.pyplot as plt
plt.rcParams.update(plt.rcParamsDefault)

class Trapezoidal(Integral):
    """Calcula la integral de una función usando la regla del trapecio compuesta.

    Attributes
    ----------
    f: callable
        Función a integrar.
    a: float
        Límite inferior de integración.
    b: float
        Límite superior de integración.
    n: int
        Número de segmentos o trapecios en los que se divide el intervalo.
    integral : float
        Resultado del cálculo de la integral.

    Methods
    -------
    graficar():
        Genera una gráfica del proceso de integración.

    Examples
    -------
    from mnspy import Trapezoidal

    def f(x):
        return (x + 1 / x) ** 2

    trap = Trapezoidal(f, 1, 2, 6)
    trap.graficar()
    print(trap.integral)
    """

    def __init__(self, f: callable, a: float, b: float, n: int = 100):
        """Constructor de la clase Trapezoidal

        Parameters
        ----------
        f: callable
            Función a integrar, f(x).
        a: float
            Límite inferior de integración.
        b: float
            Límite superior de integración.
        n: int, optional
            Número de segmentos a utilizar. Por defecto es 100.
        """
        super().__init__(f=f, a=a, b=b, n=n)
        self._calcular()

    def _calcular(self):
        """Calcula la integral por el método del trapecio compuesto.

        El resultado se almacena en el atributo `self.integral`.

        Returns
        -------
        None
        """
        h = (self._b - self._a) / self._n
        # Suma de todos los puntos intermedios
        s = sum(self._f(self._a + i * h) for i in range(1, self._n))
        # Fórmula del trapecio compuesto
        self.integral = h * (0.5 * self._f(self._a) + s + 0.5 * self._f(self._b))

    def graficar(self):
        """Genera una gráfica del proceso de integración.

        Muestra la función y los trapecios utilizados para aproximar el área.

        Returns
        -------
        None
            Muestra una gráfica de Matplotlib.
        """
        x = np.linspace(self._a, self._b, self._n + 1)
        y = self._f(x)
        plt.stem(x, y, linefmt='C2--', markerfmt='C0o', basefmt='C2-')
        # plt.plot(x, y, 'o--', color='C2', lw=1)
        plt.fill_between(x, y, color='green', alpha=0.3, label='Regla del Trapecio')
        plt.title(r'$\int{f(x)}\approx ' + str(self.integral) + '$')
        self._graficar_datos()


def main():
    """Función principal para demostración."""
    def f(x):
        return (x + 1 / x) ** 2

    trap = Trapezoidal(f, 1, 2, 6)
    trap.graficar()
    print(trap.integral)


if __name__ == '__main__':
    main()



"""Módulo de Método de Volúmenes Finitos (MVF)

Este submódulo de `mnspy` implementa el Método de Volúmenes Finitos (MVF o FVM
por sus siglas en inglés) para resolver ecuaciones diferenciales parciales,
particularmente ecuaciones de transporte (convección-difusión).

Conceptos Clave
----------------
El MVF discretiza el dominio en un conjunto de volúmenes de control (celdas)
y aplica las leyes de conservación de forma integral sobre cada uno. Es un
método robusto que garantiza la conservación de las propiedades físicas.

- **`Vertice`**: Representa un nodo o esquina en la malla.
- **`Superficie`**: Representa una cara o frontera que delimita un volumen de control.
- **`Celda`**: Representa un volumen de control, el núcleo del método donde se
  realiza el balance de flujos.
- **`VolumenFinito`**: La clase principal que gestiona la malla, ensambla y
  resuelve el sistema de ecuaciones, y realiza el post-proceso.

El módulo también incluye clases para diferentes condiciones de frontera
(`SuperficieDirichlet`, `SuperficieNeumann`, `SuperficieRobin`) y esquemas
de discretización (`Metodo`: CDS, UDS, HDS).

"""

from .mvf import Vertice, Superficie, Celda, SuperficieNeumann, SuperficieDirichlet, SuperficieRobin, Metodo, \
    es_superficie_dirichlet, es_superficie_neumann, es_superficie_robin
from .volumen_finito import VolumenFinito



from mnspy.integrales import Integral
from scipy.interpolate import interp1d
import numpy as np
import matplotlib.pyplot as plt
import sys
plt.rcParams.update(plt.rcParamsDefault)

class Simpson38(Integral):
    """Calcula la integral de una función usando la regla de Simpson 3/8 compuesta.

    Attributes
    ----------
    f: callable
        Función a integrar.
    a: float
        Límite inferior de integración.
    b: float
        Límite superior de integración.
    n: int
        Número de segmentos (debe ser un múltiplo de 3).
    integral : float
        Resultado del cálculo de la integral.

    Methods
    -------
    graficar():
        Genera una gráfica del proceso de integración.

    Examples
    -------
    from mnspy import Simpson38
    import numpy as np

    def g(x):
        return 0.2 + 25 * x - 200 * x ** 2 + 675 * x ** 3 - 900 * x ** 4 + 400 * x ** 5

    sim = Simpson38(g, 0, 0.8, 3)
    sim.graficar()
    print(sim.integral)
    """
    def __init__(self, f: callable, a: float, b: float, n: int = 99):
        """Constructor de la clase Simpson38

        Parameters
        ----------
        f: callable
            Función a integrar, f(x).
        a: float
            Límite inferior de integración.
        b: float
            Límite superior de integración.
        n: int, optional
            Número de segmentos. Debe ser un múltiplo de 3. Por defecto es 99.
        """
        super().__init__(f=f, a=a, b=b, n=n)
        if self._n % 3 != 0:
            print("Error: El número de segmentos (n) debe ser múltiplo de 3 para la regla de Simpson 3/8.")
            sys.exit()
        self._calcular()

    def _calcular(self):
        """Calcula la integral por el método de Simpson 3/8.

        El resultado se almacena en el atributo `self.integral`.

        Returns
        -------
        None
        """
        h = (self._b - self._a) / self._n
        x_vals = np.linspace(self._a, self._b, self._n + 1)
        y_vals = self._f(x_vals)

        s = y_vals[0] + y_vals[-1]
        # Suma de los puntos con i no múltiplo de 3 (multiplicados por 3)
        s += 3 * np.sum(y_vals[1:-1][(np.arange(1, self._n) % 3) != 0])
        # Suma de los puntos con i múltiplo de 3 (multiplicados por 2)
        s += 2 * np.sum(y_vals[3:-1:3])

        self.integral = (3 * h / 8) * s

    def graficar(self):
        """Genera una gráfica del proceso de integración.

        Muestra la función y las cúbicas de interpolación utilizadas por la
        regla de Simpson 3/8 para aproximar el área.

        Returns
        -------
        None
            Muestra una gráfica de Matplotlib.
        """
        x = np.linspace(self._a, self._b, self._n + 1)
        y = self._f(x)
        xvals = np.linspace(self._a, self._b, 100)
        # Se usa una interpolación cúbica para visualizar las curvas
        spl = interp1d(x, y, kind='cubic')
        y_smooth = spl(xvals)
        plt.stem(x, y, linefmt='C2--', markerfmt='C0o', basefmt='C2-')
        # plt.plot(x, y, 'o--', color='C2', lw=1)
        plt.fill_between(xvals, y_smooth, color='green', alpha=0.3, label='Regla Simpson 3/8')
        plt.title(r'$\int{f(x)}\approx ' + str(self.integral) + '$')
        self._graficar_datos()


def main():
    """Función principal para demostración."""
    def g(x):
        return 0.2 + 25 * x - 200 * x ** 2 + 675 * x ** 3 - 900 * x ** 4 + 400 * x ** 5

    sim = Simpson38(g, 0, 0.8, 3)
    sim.graficar()
    print(sim.integral)


if __name__ == '__main__':
    main()



from mnspy.ecuaciones_algebraicas_lineales import EcuacionesAlgebraicasLineales
import numpy as np


class Gauss(EcuacionesAlgebraicasLineales):
    """Clase para la implementación de la solución de un sistema de ecuaciones por el método de Gauss.

    Attributes
    ----------
    _A: ndarray
        matrix cuadrada de coeficientes del sistema
    _b: ndarray
        matrix columna de términos independientes
    _etiquetas: dict[str:list[str], str: bool]
        diccionario con etiquetas de la solución
        key: 'label' contiene una lista de etiquetas de la solución
        key: 'es_latex' un boleano que define si el string es en formato Latex o no
    _n: int
        número de filas de la matrix _A
    _m: int
        número de columnas de la matrix _A
    x: ndarray
        matrix columna de la solución del sistema de acuaciones lineales

    Methods
    -------
    _calcular():
        Soluciona el sistema de cuaciones lineales por el método de Gauss y la solución queda guardada en el
        atributo x

    solucion():
        Presenta los resultados de la solución del sistema de ecuaciones

    Examples
    -------
    from mnspy import Gauss
    import numpy as np

    A = np.matrix('3 -0.1 -0.2;0.1 7 -0.3; 0.3 -0.2 10')
    print('Matrix A:\n', A)
    b = np.matrix('7.85 ; -19.3; 71.4')
    print('Matrix b:\n', b)
    g = Gauss(A, b, pivote_parcial=True)
    print('Matrix aumentada:')
    g.mostrar_aumentada(3)
    print('solución:')
    g.solucion()
    """
    def __init__(self, A: np.ndarray, b: np.ndarray, pivote_parcial: bool = False):
        """Constructor de la clase Gauss

        Parameters
        ----------
        A: ndarray
            Matrix de los coeficientes del sistema de ecuaciones
        b: ndarray
            Vector columna de los terminos independiantes
        pivote_parcial: bool
            booleano que indica si al método se le aplica el pivote parcial o no
        """
        super().__init__(A, b)
        self._aumentada = np.hstack((self._A, self._b))
        self._pivote_parcial = pivote_parcial
        if self._n >0:
            self._calcular()

    def _calcular(self):
        """Soluciona el sistema de ecuaciones lineales por el método de Gauss y la solución queda guardada en el
        atributo x

        Returns
        -------
        None
        """
        if self._n != self._m:
            print('La matriz A debe ser cuadrada')
            return
        self.eliminacion_adelante(self._pivote_parcial)
        self.x = self.sustitucion_atras()


def main():
    A = np.array([[3, -0.1, -0.2,], [0.1, 7, -0.3], [0.3, -0.2, 10]])
    print('Matrix A:\n', A)
    b = np.array([[7.85], [-19.3], [71.4]])
    print('Matrix b:\n', b)
    g = Gauss(A, b, pivote_parcial=True)
    print('Matrix aumentada:')
    g.mostrar_aumentada(3)
    print('solución:')
    g.solucion()


if __name__ == '__main__':
    main()



from mnspy.ecuaciones_algebraicas_lineales import EcuacionesAlgebraicasLineales
import numpy as np


class Tridiagonal(EcuacionesAlgebraicasLineales):
    """Clase para la implementación de la solución de un sistema de ecuaciones por el método de Tridiagonal.

    Attributes
    ----------
    _A: ndarray
        matrix cuadrada de coeficientes del sistema
    _b: ndarray
        matrix columna de términos independientes
    _etiquetas: dict[str:list[str], str: bool]
        diccionario con etiquetas de la solución
        key: 'label' contiene una lista de etiquetas de la solución
        key: 'es_latex' un boleano que define si el string es en formato Latex o no
    _n: int
        número de filas de la matrix _A
    _m: int
        número de columnas de la matrix _A
    x: ndarray
        matrix columna de la solución del sistema de acuaciones lineales

    Methods
    -------
    _calcular():
        Soluciona el sistema de cuaciones lineales por el método de Tridiagonal y la solución queda guardada en el
        atributo x

    solucion():
        Presenta los resultados de la solución del sistema de ecuaciones

    Examples
    -------
    from mnspy import Tridiagonal
    import numpy as np

    n = 4
    e = np.zeros([n])
    f = np.zeros([n])
    g = np.zeros([n])
    for i in range(n):
        f[i] = 2.04
        if i < n - 1:
            g[i] = -1
        if i > 0:
            e[i] = -1
    r = np.array((40.8, 0.8, 0.8, 200.8))
    print('Matrix e:\n', e)
    print('Matrix f:\n', f)
    print('Matrix g:\n', g)
    print('Matrix r:\n', r)
    T = Tridiagonal(e, f, g, r)
    T.ajustar_etiquetas(['T_0', 'T_1', 'T_2', 'T_3'],
                        )
    T.mostrar_sistema()
    print('Temperaturas en grados Celsius:')
    for i in range(4):
        print(' {0:6.2f}'.format(T.x[i, 0]))
    T.solucion()
    """

    def __init__(self, e: np.ndarray, f: np.ndarray, g: np.ndarray, r: np.ndarray):
        """Constructor de la clase Tridiagonal

        Parameters
        ----------
        e: ndarray
            Vector subdiagonal de longitud n, primer elemento = 0
        f: ndarray
            Vector diagonal de longitud n.
        g: ndarray
            Vector superdiagonal de longitud n, último elemento = 0
        r: ndarray
            Vector terminos independienter de longitud n.
        """
        A = np.zeros((f.shape[0], f.shape[0]))
        # A = np.matrix(A)
        for i in range(A.shape[0]):
            for j in range(A.shape[1]):
                if i == j:
                    A[i, j] = f[i]
                elif j == i - 1:
                    A[i, j] = e[i]
                elif j == i + 1:
                    A[i, j] = g[i]
        super().__init__(A, r.reshape(-1,1))
        self._aumentada = None
        self._e = e.astype(np.double)
        self._f = f.astype(np.double)
        self._g = g.astype(np.double)
        self._r = r.astype(np.double)
        self._calcular()

    def _calcular(self):
        """Soluciona el sistema de ecuaciones lineales por el método de la Tridiagonal y la solución queda guardada
        en el atributo x

        Returns
        -------
        None
        """
        x = np.zeros([self._n])
        f = self._f.copy()
        r = self._r.copy()
        for k in range(1, self._n):
            factor = self._e[k] / f[k - 1]
            f[k] -= factor * self._g[k - 1]
            r[k] -= factor * r[k - 1]
        x[self._n - 1] = r[self._n - 1] / f[self._n - 1]
        for k in range(self._n - 2, -1, -1):
            x[k] = (r[k] - self._g[k] * x[k + 1]) / f[k]
        self.x = x.reshape(-1,1)


def main():
    n = 4
    e = np.zeros([n])
    f = np.zeros([n])
    g = np.zeros([n])
    for i in range(n):
        f[i] = 2.04
        if i < n - 1:
            g[i] = -1
        if i > 0:
            e[i] = -1
    r = np.array((40.8, 0.8, 0.8, 200.8))
    print('Matrix e:\n', e)
    print('Matrix f:\n', f)
    print('Matrix g:\n', g)
    print('Matrix r:\n', r)
    T = Tridiagonal(e, f, g, r)
    T.ajustar_etiquetas(['T_0', 'T_1', 'T_2', 'T_3'],
                        )
    T.mostrar_sistema()
    print('Temperaturas en grados Celsius:')
    for i in range(4):
        print(' {0:6.2f}'.format(T.x[i, 0]))
    T.solucion()


if __name__ == '__main__':
    main()



import numpy as np
import matplotlib.pyplot as plt
plt.rcParams.update(plt.rcParamsDefault)

class Interpolacion:
    """Clase base para los métodos de interpolación.

    Esta clase proporciona la estructura y funcionalidad común para todos los
    métodos de interpolación, como la inicialización de los puntos de datos
    (x, y) y un método base para graficar dichos puntos.

    Attributes
    ----------
    x : np.ndarray
        Array con los datos de la variable independiente.
    y : np.ndarray
        Array con los datos de la variable dependiente.
    n : int
        Número de puntos de datos.

    Methods
    -------
    _graficar_datos():
        Dibuja un gráfico de dispersión con los puntos de datos originales.

    """

    def __init__(self, x: np.ndarray, y: np.ndarray):
        """Constructor de la clase base de Interpolación.

        Parameters
        ----------
        x : np.ndarray
            Array con los datos de la variable independiente
        y : np.ndarray
            Array con los datos de la variable dependiente
        """
        self._x = x
        self._y = y
        self._n = len(self._x)
        if len(self._y) != self._n:
            raise ValueError("Los arrays 'x' e 'y' deben tener la misma longitud.")
        plt.ioff()  # deshabilitada interactividad matplotlib

    def _graficar_datos(self) -> None:
        """Dibuja los puntos de datos originales en un gráfico.

        Returns
        -------
        None
            Muestra una gráfica de Matplotlib.
        """
        plt.scatter(self._x, self._y, marker='o', c='b', lw=2, label='Datos', zorder=10)
        plt.grid()
        plt.legend()
        plt.show()



from mnspy.derivada import Derivada
import numpy as np
import matplotlib.pyplot as plt
plt.rcParams.update(plt.rcParamsDefault)

class Richardson(Derivada):
    """Calcula la derivada numérica utilizando la extrapolación de Richardson.

    Este método mejora la precisión de las fórmulas de diferencias finitas
    combinando dos estimaciones de la derivada calculadas con diferentes
    tamaños de paso (h y h/2).

    Attributes
    ----------
    f : callable
        Función a derivar.
    h : float
        Tamaño del paso base para el cálculo.
    modo : str
        Método de diferencia finita: 'adelante', 'atrás' o 'centrada'.
    orden_error : str
        Orden de error de la aproximación a mejorar.
    derivada: float
        Resultado del último cálculo de la derivada.

    Methods
    -------
    derivar(x: float):
        Calcula la derivada en el punto `x` usando extrapolación.
    graficar():
        Grafica la función y la línea tangente en un punto dado.

    Examples
    -------
    from mnspy import Richardson
    import numpy as np

    def g(x):
        return (x + 7) * (x + 2) * (x - 4) * (x - 12) / 100

    der = Derivada(g, orden='h2')
    der.derivar(2)
    print(der.derivada)

    ri = Richardson(g, orden='h2')
    ri.derivar(2)
    ri.graficar(2)
    print(ri.derivada)
    """
    def __init__(self, f: callable, n: int = 1, h: float = 1e-3, orden: str = 'h', modo: str = 'adelante'):
        """Constructor de la clase Richardson.

        Parameters
        ----------
        f : callable
            Función a derivar, f(x).
        n : int, optional
            Grado de la derivada (1, 2, 3 o 4). Por defecto es 1.
        h : float, optional
            Tamaño del paso base. Por defecto es 1e-3.
        orden : str, optional
            Orden de error de la aproximación.
            - Para modo 'adelante' y 'atrás': 'h' o 'h2'.
            - Para modo 'centrada': 'h2' o 'h4'.
            Por defecto es 'h'.
        modo : str, optional
            Método de diferencia finita a utilizar: 'adelante', 'atrás' o 'centrada'.
            Por defecto es 'adelante'.
        """
        super().__init__(f, n, h, orden, modo)

    def derivar(self, x: float):
        """Calcula la derivada en `x` usando la extrapolación de Richardson.

        Parameters
        ----------
        x : float
            Punto en el que se evaluará la derivada.
        """
        # Guarda el h original
        h_original = self._h

        # Calcula D(h/2)
        self._h = h_original / 2.0
        super().derivar(x)
        d_h2 = self.derivada

        # Calcula D(h)
        self._h = h_original
        super().derivar(x)
        d_h1 = self.derivada

        # Aplica la fórmula de extrapolación de Richardson para O(h^2)
        self.derivada = (4 * d_h2 - d_h1) / 3

    def graficar(self, x: float, x_min: float = None, x_max: float = None, delta: float = 10) -> None:
        """Grafica la función y la línea tangente mejorada por Richardson.

        Parameters
        ----------
        x: float
            posición en x en que se dibujará la derivada
        x_min: float
            valor mínimo de la gráfica en x
        x_max: float
            valor máximo de la gráfica en x
        delta: float, optional
            Rango de la gráfica en x si `x_min` y `x_max` no se especifican.

        Returns
        -------
        None
            Muestra una gráfica de Matplotlib.
        """
        n = self._n
        self._n = 1
        self.derivar(x)
        self._n = n
        if x_min is None:
            x_min = x - delta
        if x_max is None:
            x_max = x + delta
        if self._orden == 'h':
            orden = r'$\mathcal{O}(h^2)$'
        elif self._orden == 'h2':
            orden = r'$\mathcal{O}(h^4)$'
        else:
            orden = r'$\mathcal{O}(h^6)$'
        y = self._f(x)
        x_list = np.linspace(x_min, x_max, 100)
        y_list = self._f(x_list)
        plt.scatter(x, y, c='r', lw=2, label=f'Punto ({x}, {self._f(x):.4f})')
        plt.plot(x_list, y_list, linestyle='-', c='b', lw=2, label='$f(x)$')
        plt.title(f'Derivada = {self.derivada:.8f}')
        plt.suptitle(
            f'Derivada Richardson, $h_1$={self._h}, $h_2$={self._h / 2}, modo={self._modo}, nuevo orden={orden}')
        plt.axline((x, y), slope=self.derivada, linestyle='dashed', c='r', lw=2,
                   label='Derivada')
        plt.grid()
        plt.legend()
        plt.show()


def main():
    """Función principal para demostración."""
    def g(x):
        return (x + 7) * (x + 2) * (x - 4) * (x - 12) / 100

    der = Derivada(g, orden='h2')
    der.derivar(2)
    print(der.derivada)

    ri = Richardson(g, orden='h2')
    ri.derivar(2)
    ri.graficar(2)
    print(ri.derivada)


if __name__ == '__main__':
    main()



"""Módulo Ecuaciones Diferenciales Ordinarias (EDOs)

Este módulo de `mnspy` proporciona una colección de métodos numéricos para
resolver problemas de valor inicial para ecuaciones diferenciales ordinarias (EDOs).

Los métodos implementados son adecuados para resolver EDOs de primer orden de la
forma `dy/dx = f(x, y)`.

Clases Disponibles
------------------
- `EcuacionesDiferencialesOrdinarias`:
    Clase base abstracta de la que heredan todos los solucionadores de EDOs.
- `Euler`:
    Implementación del método de Euler, el método explícito más simple.
- `Heun`:
    Implementación del método de Heun, un método predictor-corrector simple.
- `PuntoMedio`:
    Implementación del método del punto medio (RK2 modificado).
- `RungeKutta`:
    Implementación de los métodos clásicos de Runge-Kutta de órdenes 2, 3, 4 y 5.

"""
from .ecuaciones_diferenciales_ordinarias import EcuacionesDiferencialesOrdinarias
from .euler import Euler
from .heun import Heun
from .punto_medio import PuntoMedio
from .runge_kutta import RungeKutta
